"""HTML escape test cases."""
# pylint: disable=missing-class-docstring
from unittest import TestCase
from unittest import skipIf

from typing import NamedTuple
from typing import Dict
from typing import List

try:
    import markupsafe
except ImportError:
    markupsafe = None

try:
    from markupsafe import Markup
    from markupsafe import escape
except ImportError:
    from liquid.exceptions import Markup  # type: ignore
    from liquid.exceptions import escape  # type: ignore

from liquid import Environment
from liquid.exceptions import Error


class Case(NamedTuple):
    """Table driven test case helper."""

    description: str
    template: str
    context: Dict
    expect: str


class SafeHTMLDrop:
    def __init__(self, somelist: List[object]):
        self.items = somelist

    def __str__(self):
        return "SafeHTMLDrop"

    def __html__(self):
        lis = "\n".join(f"<li>{item}</li>" for item in self.items)
        return f"<ul>\n{lis}\n</ul>"


@skipIf(markupsafe is None, "this test requires markupsafe")
class AutoescapeTestCase(TestCase):
    def test_do_not_escape_output(self):
        """Test that we can turn off auto-escaping."""
        tests = [
            Case(
                description="script from context",
                template=r"{{ name }}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect='<script>alert("XSS!");</script>',
            ),
            Case(
                description="__html__ is ignored",
                template=r"{{ foo }}",
                context={"foo": SafeHTMLDrop([1, 2, 3])},
                expect="SafeHTMLDrop",
            ),
        ]

        env = Environment(autoescape=False)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_escape_output(self):
        """Test that we can automatically escape output statements."""
        tests = [
            Case(
                description="script from context",
                template=r"{{ name }}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="&lt;script&gt;alert(&#34;XSS!&#34;);&lt;/script&gt;",
            ),
            Case(
                description="script from context inside block",
                template=r"{% if true %}{{ name }}{% endif %}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="&lt;script&gt;alert(&#34;XSS!&#34;);&lt;/script&gt;",
            ),
            Case(
                description="script from context inside block with literal",
                template=r"{% if true %}<br>{{ name }}{% endif %}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="<br>&lt;script&gt;alert(&#34;XSS!&#34;);&lt;/script&gt;",
            ),
            Case(
                description="capture with HTML literals",
                template=(
                    r"{% capture foo %}"
                    r'<p class="foo">'
                    r"{{ name }}"
                    r"</p>"
                    r"{% endcapture %}"
                    r"{{ foo }}"
                ),
                context={"name": '<script>alert("XSS!");</script>'},
                expect=(
                    '<p class="foo">'
                    "&lt;script&gt;alert(&#34;XSS!&#34;);&lt;/script&gt;"
                    "</p>"
                ),
            ),
            Case(
                description="drops with __html__ are safe",
                template=(r"{{ foo }}"),
                context={"foo": SafeHTMLDrop([1, 2, 3])},
                expect="<ul>\n<li>1</li>\n<li>2</li>\n<li>3</li>\n</ul>",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_escape_string_filters(self):
        """Test that string filters are autoescaped."""
        tests = [
            Case(
                description="upcase",
                template=r"{{ name | upcase }}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="&lt;SCRIPT&gt;ALERT(&#34;XSS!&#34;);&lt;/SCRIPT&gt;",
            ),
            Case(
                description="upcase safe",
                template=r"{{ name | upcase }}",
                context={"name": Markup('<script>alert("XSS!");</script>')},
                expect='<SCRIPT>ALERT("XSS!");</SCRIPT>',
            ),
            Case(
                description="downcase",
                template=r"{{ name | downcase }}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="&lt;script&gt;alert(&#34;xss!&#34;);&lt;/script&gt;",
            ),
            Case(
                description="downcase safe",
                template=r"{{ name | downcase }}",
                context={"name": Markup('<script>alert("XSS!");</script>')},
                expect='<script>alert("xss!");</script>',
            ),
            Case(
                description="capitalize",
                template=r"{{ name | capitalize }}",
                context={"name": '<script>alert("XSS!");</script>'},
                expect="&lt;script&gt;alert(&#34;xss!&#34;);&lt;/script&gt;",
            ),
            Case(
                description="capitalize safe",
                template=r"{{ name | capitalize }}",
                context={"name": Markup('<script>alert("XSS!");</script>')},
                expect='<script>alert("xss!");</script>',
            ),
            Case(
                description="append unsafe left value and unsafe argument",
                template=r"{{ some | append: other }}",
                context={"some": "<br>", "other": "<hr>"},
                expect="&lt;br&gt;&lt;hr&gt;",
            ),
            Case(
                description="append safe left value and unsafe argument",
                template=r"{{ some | append: other }}",
                context={"some": Markup("<br>"), "other": "<hr>"},
                expect="<br>&lt;hr&gt;",
            ),
            Case(
                description="append safe left value and safe argument",
                template=r"{{ some | append: other }}",
                context={"some": Markup("<br>"), "other": Markup("<hr>")},
                expect="<br><hr>",
            ),
            Case(
                description="lstrip",
                template=r"{{ some | lstrip }}",
                context={"some": "   <br>"},
                expect="&lt;br&gt;",
            ),
            Case(
                description="lstrip safe",
                template=r"{{ some | lstrip }}",
                context={"some": Markup("   <br>")},
                expect="<br>",
            ),
            Case(
                description="newline to BR",
                template=r"{{ some | newline_to_br }}",
                context={"some": "<em>hello</em>\n<b>goodbye</b>"},
                expect="&lt;em&gt;hello&lt;/em&gt;<br />\n&lt;b&gt;goodbye&lt;/b&gt;",
            ),
            Case(
                description="newline to BR safe",
                template=r"{{ some | newline_to_br }}",
                context={"some": Markup("<em>hello</em>\n<b>goodbye</b>")},
                expect="<em>hello</em><br />\n<b>goodbye</b>",
            ),
            Case(
                description="newline to BR chained filter",
                template=r"{{ some | newline_to_br | upcase }}",
                context={"some": "<em>hello</em>\n<b>goodbye</b>"},
                expect="&LT;EM&GT;HELLO&LT;/EM&GT;<BR />\n&LT;B&GT;GOODBYE&LT;/B&GT;",
            ),
            Case(
                description="newline to BR safe chained filter",
                template=r"{{ some | newline_to_br | upcase }}",
                context={"some": Markup("<em>hello</em>\n<b>goodbye</b>")},
                expect="<EM>HELLO</EM><BR />\n<B>GOODBYE</B>",
            ),
            Case(
                description="prepend unsafe left value and unsafe argument",
                template=r"{{ some | prepend: other }}",
                context={"some": "<br>", "other": "<hr>"},
                expect="&lt;hr&gt;&lt;br&gt;",
            ),
            Case(
                description="prepend safe left value and unsafe argument",
                template=r"{{ some | prepend: other }}",
                context={"some": Markup("<br>"), "other": "<hr>"},
                expect="&lt;hr&gt;<br>",
            ),
            Case(
                description="prepend safe left value and safe argument",
                template=r"{{ some | prepend: other }}",
                context={"some": Markup("<br>"), "other": Markup("<hr>")},
                expect="<hr><br>",
            ),
            Case(
                description="remove unsafe left value and unsafe argument",
                template=r"{{ some | remove: other }}",
                context={"some": "<br><p>hello</p><br>", "other": "<br>"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="remove safe left value and unsafe argument",
                template=r"{{ some | remove: other }}",
                context={"some": Markup("<br><p>hello</p><br>"), "other": "<br>"},
                expect="<br><p>hello</p><br>",
            ),
            Case(
                description="remove safe left value and safe argument",
                template=r"{{ some | remove: other }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "other": Markup("<br>"),
                },
                expect="<p>hello</p>",
            ),
            Case(
                description="remove first unsafe left value and unsafe argument",
                template=r"{{ some | remove_first: other }}",
                context={"some": "<br><p>hello</p><br>", "other": "<br>"},
                expect="&lt;p&gt;hello&lt;/p&gt;&lt;br&gt;",
            ),
            Case(
                description="remove first safe left value and unsafe argument",
                template=r"{{ some | remove_first: other }}",
                context={"some": Markup("<br><p>hello</p><br>"), "other": "<br>"},
                expect="<br><p>hello</p><br>",
            ),
            Case(
                description="remove first safe left value and safe argument",
                template=r"{{ some | remove_first: other }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "other": Markup("<br>"),
                },
                expect="<p>hello</p><br>",
            ),
            Case(
                description="replace unsafe left value and unsafe arguments",
                template=r"{{ some | replace: seq, sub }}",
                context={
                    "some": "<br><p>hello</p><br>",
                    "seq": "<br>",
                    "sub": "<hr>",
                },
                expect="&lt;hr&gt;&lt;p&gt;hello&lt;/p&gt;&lt;hr&gt;",
            ),
            Case(
                description="replace safe left value and unsafe arguments",
                template=r"{{ some | replace: seq, sub }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "seq": "<br>",
                    "sub": "<hr>",
                },
                expect="<br><p>hello</p><br>",
            ),
            Case(
                description="replace safe left value and safe arguments",
                template=r"{{ some | replace: seq, sub }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "seq": Markup("<br>"),
                    "sub": Markup("<hr>"),
                },
                expect="<hr><p>hello</p><hr>",
            ),
            Case(
                description="replace first - unsafe left value and unsafe arguments",
                template=r"{{ some | replace_first: seq, sub }}",
                context={
                    "some": "<br><p>hello</p><br>",
                    "seq": "<br>",
                    "sub": "<hr>",
                },
                expect="&lt;hr&gt;&lt;p&gt;hello&lt;/p&gt;&lt;br&gt;",
            ),
            Case(
                description="replace first - safe left value and unsafe arguments",
                template=r"{{ some | replace_first: seq, sub }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "seq": "<br>",
                    "sub": "<hr>",
                },
                expect="<br><p>hello</p><br>",
            ),
            Case(
                description="replace first - safe left value and safe arguments",
                template=r"{{ some | replace_first: seq, sub }}",
                context={
                    "some": Markup("<br><p>hello</p><br>"),
                    "seq": Markup("<br>"),
                    "sub": Markup("<hr>"),
                },
                expect="<hr><p>hello</p><br>",
            ),
            Case(
                description="slice",
                template=r"{{ some | slice: 4, 12 }}",
                context={"some": "<br><p>hello</p><br>"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="slice safe",
                template=r"{{ some | slice: 4, 12 }}",
                context={"some": Markup("<br><p>hello</p><br>")},
                expect="<p>hello</p>",
            ),
            Case(
                description="split unsafe left value and unsafe argument",
                template=r"{{ some | split: other }}",
                context={"some": "<p>hello</p><br><p>goodbye</p>", "other": "<br>"},
                expect="&lt;p&gt;hello&lt;/p&gt;&lt;p&gt;goodbye&lt;/p&gt;",
            ),
            Case(
                description="split safe left value and unsafe argument",
                template=r"{{ some | split: other }}",
                context={
                    "some": Markup("<p>hello</p><br><p>goodbye</p>"),
                    "other": "<br>",
                },
                expect="<p>hello</p><p>goodbye</p>",
            ),
            Case(
                description="split safe left value and safe argument",
                template=r"{{ some | split: other }}",
                context={
                    "some": Markup("<p>hello</p><br><p>goodbye</p>"),
                    "other": Markup("<br>"),
                },
                expect="<p>hello</p><p>goodbye</p>",
            ),
            Case(
                description="strip",
                template=r"{{ some | strip }}",
                context={"some": "\n<p>hello</p>  \n"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="strip safe",
                template=r"{{ some | strip }}",
                context={"some": Markup("\n<p>hello</p>  \n")},
                expect="<p>hello</p>",
            ),
            Case(
                description="right strip",
                template=r"{{ some | rstrip }}",
                context={"some": "\n<p>hello</p>  \n"},
                expect="\n&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="right strip safe",
                template=r"{{ some | rstrip }}",
                context={"some": Markup("\n<p>hello</p>  \n")},
                expect="\n<p>hello</p>",
            ),
            Case(
                description="strip html",
                template=r"{{ some | strip_html }}",
                context={"some": "<p>hello</p>"},
                expect="hello",
            ),
            Case(
                description="strip html safe",
                template=r"{{ some | strip_html }}",
                context={"some": Markup("<p>hello</p>")},
                expect="hello",
            ),
            Case(
                description="strip newlines",
                template=r"{{ some | strip_newlines }}",
                context={"some": "\n<p>hello</p>  \n"},
                expect="&lt;p&gt;hello&lt;/p&gt;  ",
            ),
            Case(
                description="strip newlines safe",
                template=r"{{ some | strip_newlines }}",
                context={"some": Markup("\n<p>hello</p>  \n")},
                expect="<p>hello</p>  ",
            ),
            Case(
                description="truncate",
                template=r"{{ some | truncate: 10, '' }}",
                context={"some": "<p>hello</p>"},
                expect="&lt;p&gt;hello&lt;/",
            ),
            Case(
                description="truncate safe",
                template=r"{{ some | truncate: 10, '' }}",
                context={"some": Markup("<p>hello</p>")},
                expect="&lt;p&gt;hello&lt;/",
            ),
            Case(
                description="truncate words",
                template=r"{{ some | truncatewords: 3 }}",
                context={"some": "<em>Ground</em> control to Major Tom."},
                expect="&lt;em&gt;Ground&lt;/em&gt; control to...",
            ),
            Case(
                description="truncate words safe",
                template=r"{{ some | truncatewords: 3 }}",
                context={"some": Markup("<em>Ground</em> control to Major Tom.")},
                expect="&lt;em&gt;Ground&lt;/em&gt; control to...",
            ),
            Case(
                description="URL encode",
                template=r"{{ some | url_encode }}",
                context={"some": "<p>hello</p>"},
                expect=r"%3Cp%3Ehello%3C%2Fp%3E",
            ),
            Case(
                description="URL encode safe",
                template=r"{{ some | url_encode }}",
                context={"some": Markup("<p>hello</p>")},
                expect=r"%3Cp%3Ehello%3C%2Fp%3E",
            ),
            Case(
                description="URL decode",
                template=r"{{ some | url_decode }}",
                context={"some": r"%3Cp%3Ehello%3C%2Fp%3E"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="URL decode safe",
                template=r"{{ some | url_decode }}",
                context={"some": Markup(r"%3Cp%3Ehello%3C%2Fp%3E")},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="liquid escape",
                template=r"{{ some | escape }}",
                context={"some": "<p>hello</p>"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="liquid escape markup",
                template=r"{{ some | escape }}",
                context={"some": Markup("<p>hello</p>")},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="escape once",
                template=r"{{ some | escape_once }}",
                context={"some": "&lt;p&gt;test&lt;/p&gt;<p>test</p>"},
                expect="&lt;p&gt;test&lt;/p&gt;&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="escape once markup",
                template=r"{{ some | escape_once }}",
                context={"some": Markup("&lt;p&gt;test&lt;/p&gt;<p>test</p>")},
                expect="&lt;p&gt;test&lt;/p&gt;&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="escape __html__",
                template=r"{{ some | escape }}",
                context={"some": SafeHTMLDrop([1, 2, 3])},
                expect="SafeHTMLDrop",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_autoescape_array_filters(self):
        """Test that array filters are autoescaped."""
        tests = [
            Case(
                description="join unsafe iterable and default separator",
                template=r"{{ foo | join }}",
                context={"foo": ["<p>hello</p>", "<p>goodbye</p>"]},
                expect="&lt;p&gt;hello&lt;/p&gt; &lt;p&gt;goodbye&lt;/p&gt;",
            ),
            Case(
                description="join unsafe iterable and unsafe separator",
                template=r"{{ foo | join: bar }}",
                context={"foo": ["<p>hello</p>", "<p>goodbye</p>"], "bar": "<hr>"},
                expect="&lt;p&gt;hello&lt;/p&gt;&lt;hr&gt;&lt;p&gt;goodbye&lt;/p&gt;",
            ),
            Case(
                description="join safe iterable and default separator",
                template=r"{{ foo | join }}",
                context={"foo": [Markup("<p>hello</p>"), Markup("<p>goodbye</p>")]},
                expect="<p>hello</p> <p>goodbye</p>",
            ),
            Case(
                description="join safe iterable and unsafe separator",
                template=r"{{ foo | join: bar }}",
                context={
                    "foo": [Markup("<p>hello</p>"), Markup("<p>goodbye</p>")],
                    "bar": "<hr>",
                },
                expect="&lt;p&gt;hello&lt;/p&gt;&lt;hr&gt;&lt;p&gt;goodbye&lt;/p&gt;",
            ),
            Case(
                description="join safe iterable and safe separator",
                template=r"{{ foo | join: bar }}",
                context={
                    "foo": [Markup("<p>hello</p>"), Markup("<p>goodbye</p>")],
                    "bar": Markup("<hr>"),
                },
                expect="<p>hello</p><hr><p>goodbye</p>",
            ),
            Case(
                description="join mixed iterable and safe separator",
                template=r"{{ foo | join: bar }}",
                context={
                    "foo": [Markup("<p>hello</p>"), "<p>goodbye</p>"],
                    "bar": Markup("<hr>"),
                },
                expect="<p>hello</p><hr>&lt;p&gt;goodbye&lt;/p&gt;",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_autoescape_misc_filters(self):
        """Test that misc filters are autoescaped."""
        tests = [
            Case(
                description="date",
                template=r"{{ foo | date: bar }}",
                context={"foo": "March 14, 2016", "bar": r"%b %d, %y<hr>"},
                expect="Mar 14, 16&lt;hr&gt;",
            ),
            Case(
                description="date with literal format string",
                template=r"{{ foo | date: '%b %d, %y<br>' }}",
                context={"foo": "March 14, 2016"},
                expect="Mar 14, 16<br>",
            ),
            Case(
                description="date with safe format string from context",
                template=r"{{ foo | date: bar }}",
                context={"foo": "March 14, 2016", "bar": Markup(r"%b %d, %y<br>")},
                expect="Mar 14, 16<br>",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_safe_filter(self):
        """Test the safe filter."""
        tests = [
            Case(
                description="safe from unsafe",
                template=r"<p>Hello, {{ you | safe }}</p>",
                context={"you": "<em>World!</em>"},
                expect="<p>Hello, <em>World!</em></p>",
            ),
            Case(
                description="safe from safe",
                template=r"<p>Hello, {{ you | safe }}</p>",
                context={"you": Markup("<em>World!</em>")},
                expect="<p>Hello, <em>World!</em></p>",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_safe_filter_with_no_autoescape(self):
        """Test that the safe filter is a no-op if autoescape is disabled."""
        env = Environment(autoescape=False)
        template = env.from_string(r"{{ '<p>Hello</p>' | safe }}")
        result = template.render()
        self.assertEqual(result, "<p>Hello</p>")

    def test_tablerow_tag(self):
        """Test that the tablerow tag behaves well with autoescape."""
        tests = [
            Case(
                description="no markup",
                template=(
                    r"{% tablerow i in (1..4) cols:2 %}"
                    r"{{ i }} {{ tablerowloop.col_first }}"
                    r"{% endtablerow %}"
                ),
                context={},
                expect=(
                    '<tr class="row1">\n'
                    '<td class="col1">1 true</td>'
                    '<td class="col2">2 false</td>'
                    "</tr>\n"
                    '<tr class="row2">'
                    '<td class="col1">3 true</td>'
                    '<td class="col2">4 false</td>'
                    "</tr>\n"
                ),
            ),
            Case(
                description="markup in loop block",
                template=(
                    r"{% tablerow i in (1..4) cols:2 %}"
                    r"<b>{{ i }}</b> {{ tablerowloop.col_first }}"
                    r"{% endtablerow %}"
                ),
                context={},
                expect=(
                    '<tr class="row1">\n'
                    '<td class="col1"><b>1</b> true</td>'
                    '<td class="col2"><b>2</b> false</td>'
                    "</tr>\n"
                    '<tr class="row2">'
                    '<td class="col1"><b>3</b> true</td>'
                    '<td class="col2"><b>4</b> false</td>'
                    "</tr>\n"
                ),
            ),
            Case(
                description="unsafe markup from iterable",
                template=(
                    r"{% tablerow tag in collection.tags %}"
                    r"{{ tag }}"
                    r"{% endtablerow %}"
                ),
                expect=(
                    '<tr class="row1">\n'
                    '<td class="col1">&lt;b&gt;tag1&lt;/b&gt;</td>'
                    '<td class="col2">tag2</td>'
                    '<td class="col3">tag3</td>'
                    '<td class="col4">tag4</td>'
                    "</tr>\n"
                ),
                context={
                    "collection": {
                        "tags": [
                            "<b>tag1</b>",
                            "tag2",
                            "tag3",
                            "tag4",
                        ]
                    }
                },
            ),
            Case(
                description="safe markup from iterable",
                template=(
                    r"{% tablerow tag in collection.tags %}"
                    r"{{ tag }}"
                    r"{% endtablerow %}"
                ),
                expect=(
                    '<tr class="row1">\n'
                    '<td class="col1"><b>tag1</b></td>'
                    '<td class="col2">tag2</td>'
                    '<td class="col3">tag3</td>'
                    '<td class="col4">tag4</td>'
                    "</tr>\n"
                ),
                context={
                    "collection": {
                        "tags": [
                            Markup("<b>tag1</b>"),
                            "tag2",
                            "tag3",
                            "tag4",
                        ]
                    }
                },
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)

    def test_liquid_tag(self):
        """Test that the liquid tag behaves well with autoescape."""
        tests = [
            Case(
                description="echo markup from context",
                template="\n".join(
                    [
                        r"{% liquid ",
                        r"    echo foo",
                        r"%}",
                    ]
                ),
                context={"foo": "<em>hello</em>"},
                expect="&lt;em&gt;hello&lt;/em&gt;",
            ),
            Case(
                description="echo unsafe markup from string literal",
                template="\n".join(
                    [
                        r"{% liquid ",
                        r"    echo '<em>hello</em>'",
                        r"%}",
                    ]
                ),
                context={},
                expect="<em>hello</em>",
            ),
            Case(
                description="echo safe markup from context",
                template="\n".join(
                    [
                        r"{% liquid ",
                        r"    echo foo",
                        r"%}",
                    ]
                ),
                context={"foo": Markup("<em>hello</em>")},
                expect="<em>hello</em>",
            ),
        ]

        env = Environment(autoescape=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)


@skipIf(markupsafe is not None, "this tests exceptions raise by the lack of markupsafe")
class DummyMarkupSafeTestCase(TestCase):
    def test_dummy_escape(self):
        """Test that the dummy definition of escape raises an exception."""
        with self.assertRaises(Error):
            escape("foo")

    def test_dummy_markup(self):
        """Test that the dummy definition of Markup raises an exception."""
        with self.assertRaises(Error):
            Markup("foo")



# flake8: noqa
# pylint: disable=useless-import-alias,missing-module-docstring
from .cache import LRUCache as LRUCache
from .html import strip_tags as strip_tags
from .text import truncate_chars as truncate_chars
from .text import truncate_words as truncate_words



# pylint: disable=missing-module-docstring
from .if_not import IfNotTag
from .if_expressions import InlineIfAssignTag
from .if_expressions import InlineIfAssignTagWithParens
from .if_expressions import InlineIfEchoTag
from .if_expressions import InlineIfEchoTagWithParens
from .if_expressions import InlineIfStatement
from .if_expressions import InlineIfStatementWithParens
from .macro import MacroTag
from .macro import CallTag
from ._with import WithTag

__all__ = (
    "CallTag",
    "MacroTag",
    "IfNotTag",
    "InlineIfAssignTag",
    "InlineIfAssignTagWithParens",
    "InlineIfEchoTag",
    "InlineIfEchoTagWithParens",
    "InlineIfStatement",
    "InlineIfStatementWithParens",
    "WithTag",
)



"""Test cases for rendering non-standard "extra" tags."""
import asyncio
import unittest
from typing import List

from liquid import golden
from liquid.environment import Environment
from liquid.golden.case import Case
from liquid.loaders import DictLoader
from liquid.template import BoundTemplate

from liquid.extra import IfNotTag
from liquid.extra import WithTag
from liquid.extra import add_inline_expression_tags
from liquid.extra import add_extended_inline_expression_tags


class BaseRenderTestCase(unittest.TestCase):
    """Base class for test cases that render synchronously and asynchronously."""

    def setUp(self) -> None:
        self.partials = {}
        self.loader = DictLoader(self.partials)
        self.env = Environment(loader=self.loader)

    def _test(self, test_cases: List[Case]) -> None:
        for case in test_cases:
            self.partials.update(case.partials)

            with self.subTest(msg=case.description):
                template = self.env.from_string(case.template, globals=case.globals)
                result = template.render()
                self.assertEqual(result, case.expect)

        async def coro(template: BoundTemplate):
            return await template.render_async()

        for case in test_cases:
            self.partials.update(case.partials)

            with self.subTest(msg=case.description, asynchronous=True):
                template = self.env.from_string(case.template, globals=case.globals)
                result = asyncio.run(coro(template))
                self.assertEqual(result, case.expect)


class RenderIfNotTagTestCase(BaseRenderTestCase):
    """Test cases for non-standard `if` tag."""

    def setUp(self) -> None:
        super().setUp()
        self.env.add_tag(IfNotTag)

    def test_render_standard_if_tag(self) -> None:
        """Test that the `if (not)` tag renders standard `if` tags."""
        self._test(golden.if_tag.cases)

    def test_render_non_standard_if_tag(self) -> None:
        """Test that we can render `if` tags with logical `not` and grouping
        with parentheses."""
        test_cases = [
            Case(
                description="not false",
                template=r"{% if not false %}foo{% endif %}",
                expect="foo",
            ),
            Case(
                description="not true",
                template=r"{% if not true %}foo{% endif %}",
                expect="",
            ),
            Case(
                description="literal boolean filter",
                template=r"{{ false | default: true }}",
                expect="true",
            ),
            Case(
                description="not comparison to empty",
                template=r"{% if not '' == empty %}foo{% endif %}",
                expect="",
            ),
            Case(
                description="not contains",
                template=r"{% if not foo contains 'z' %}bar{% endif %}",
                expect="bar",
                globals={"foo": ["a", "b", "c"]},
            ),
            Case(
                description="and not",
                template=r"{% if not foo and not bar %}hello{% endif %}",
                expect="hello",
                globals={"foo": False, "bar": False},
            ),
            Case(
                description="true and not",
                template=r"{% if foo and not bar %}hello{% endif %}",
                expect="hello",
                globals={"foo": True, "bar": False},
            ),
            Case(
                description="not equals",
                template=r"{% if not foo == True %}hello{% endif %}",
                expect="hello",
                globals={"foo": False},
            ),
            Case(
                description="not not equals False",
                template=r"{% if not foo != true %}hello{% endif %}",
                expect="",
                globals={"foo": False},
            ),
            Case(
                description="not not equals true",
                template=r"{% if not foo != true %}hello{% endif %}",
                expect="hello",
                globals={"foo": True},
            ),
            Case(
                description="not contains with parens",
                template=r"{% if not (foo contains 'z') %}bar{% endif %}",
                expect="bar",
                globals={"foo": ["a", "b", "c"]},
            ),
        ]
        self._test(test_cases)


class RenderInlineIfTestCase(BaseRenderTestCase):
    """Test cases for rendering inline `if` tags and output statements."""

    def setUp(self) -> None:
        super().setUp()
        add_inline_expression_tags(self.env)

    def test_render_inline_if_output_statement(self) -> None:
        """Test that we can render output statements with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal with true condition",
                template=r"{{ 'hello' if true }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{{ 'hello' if false }}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early  filter",
                template=r"{{ 'hello' | upcase if true }}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{{ 'hello' if false else 'goodbye' }}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=r"{{ greeting if settings.foo else 'bar' || upcase }}",
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)

    def test_render_inline_if_assignment_tags(self) -> None:
        """Test that we can render `assign` tags with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal",
                template=r"{% assign foo = 'hello' %}{{ foo }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with true condition",
                template=r"{% assign foo = 'hello' if true %}{{ foo }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{% assign foo = 'hello' if false %}{{ foo }}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early  filter",
                template=r"{% assign foo = 'hello' | upcase if true %}{{ foo }}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{% assign foo = 'hello' if false else 'goodbye' %}{{ foo }}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=(
                    r"{% assign foo = greeting if settings.foo else 'bar' || upcase %}"
                    r"{{ foo }}"
                ),
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)

    def test_render_inline_if_echo_tags(self) -> None:
        """Test that we can render `echo` tags with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal",
                template=r"{% echo 'hello' %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with true condition",
                template=r"{% echo 'hello' if true %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{% echo 'hello' if false %}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early  filter",
                template=r"{% echo 'hello' | upcase if true %}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{% echo 'hello' if false else 'goodbye' %}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=(r"{% echo greeting if settings.foo else 'bar' || upcase %}"),
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)


class RenderInlineIfWithParensTestCase(BaseRenderTestCase):
    """Test cases for rendering inline `if` tags that support logical `not`
    and grouping terms with parentheses."""

    def setUp(self) -> None:
        super().setUp()
        add_extended_inline_expression_tags(self.env)

    def test_render_inline_if_output_statement(self) -> None:
        """Test that we can render output statements with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal with true condition",
                template=r"{{ 'hello' if true }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{{ 'hello' if false }}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early  filter",
                template=r"{{ 'hello' | upcase if true }}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{{ 'hello' if false else 'goodbye' }}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="negate condition",
                template=r"{{ 'hello' if not false else 'goodbye' }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="group condition terms",
                template=r"{{ 'hello' if (false and true) or true }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=r"{{ greeting if settings.foo else 'bar' || upcase }}",
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)

    def test_render_inline_if_assignment_tags(self) -> None:
        """Test that we can render `assign` tags with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal",
                template=r"{% assign foo = 'hello' %}{{ foo }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with true condition",
                template=r"{% assign foo = 'hello' if true %}{{ foo }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{% assign foo = 'hello' if false %}{{ foo }}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early  filter",
                template=r"{% assign foo = 'hello' | upcase if true %}{{ foo }}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{% assign foo = 'hello' if false else 'goodbye' %}{{ foo }}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="negate condition",
                template=r"{% assign foo = 'hello' if not false %}{{ foo }}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="group condition terms",
                template=(
                    r"{% assign foo = 'hello' if (false and true) or true %}"
                    r"{{ foo }}"
                ),
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=(
                    r"{% assign foo = greeting if settings.foo else 'bar' || upcase %}"
                    r"{{ foo }}"
                ),
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)

    def test_render_inline_if_echo_tags(self) -> None:
        """Test that we can render `echo` tags with inline `if` expressions."""
        test_cases = [
            Case(
                description="string literal",
                template=r"{% echo 'hello' %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with true condition",
                template=r"{% echo 'hello' if true %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="default to undefined",
                template=r"{% echo 'hello' if false %}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="early filter",
                template=r"{% echo 'hello' | upcase if true %}",
                expect="HELLO",
                globals={},
                partials={},
            ),
            Case(
                description="string literal with false condition and alternative",
                template=r"{% echo 'hello' if false else 'goodbye' %}",
                expect="goodbye",
                globals={},
                partials={},
            ),
            Case(
                description="negate condition",
                template=r"{% echo 'hello' if not false else 'goodbye' %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="group condition terms",
                template=r"{% echo 'hello' if (false and true) or true %}",
                expect="hello",
                globals={},
                partials={},
            ),
            Case(
                description="object and condition from context with tail filter",
                template=(r"{% echo greeting if settings.foo else 'bar' || upcase %}"),
                expect="HELLO",
                globals={"settings": {"foo": True}, "greeting": "hello"},
                partials={},
            ),
        ]
        self._test(test_cases)


class RenderWithTagTestCase(BaseRenderTestCase):
    """Test cases for non-standard `with` tag."""

    def setUp(self) -> None:
        super().setUp()
        self.env.add_tag(WithTag)

    def test_render_with_tag(self) -> None:
        """Test that we can render a `with` tag."""
        test_cases = [
            Case(
                description="block scoped variable",
                template=r"{{ x }}{% with x: 'foo' %}{{ x }}{% endwith %}{{ x }}",
                expect="foo",
            ),
            Case(
                description="block scoped alias",
                template=(
                    r"{% with p: collection.products.first %}"
                    r"{{ p.title }}"
                    r"{% endwith %}"
                    r"{{ p.title }}"
                    r"{{ collection.products.first.title }}"
                ),
                expect="A ShoeA Shoe",
                globals={"collection": {"products": [{"title": "A Shoe"}]}},
            ),
            Case(
                description="multiple block scoped variables",
                template=(
                    r"{% with a: 1, b: 3.4 %}"
                    r"{{ a }} + {{ b }} = {{ a | plus: b }}"
                    r"{% endwith %}"
                ),
                expect="1 + 3.4 = 4.4",
            ),
        ]
        self._test(test_cases)



# pylint: disable=missing-class-docstring missing-module-docstring
import unittest

from typing import NamedTuple
from typing import Union

from liquid.expression import Expression
from liquid.expression import BooleanExpression
from liquid.expression import IdentifierPathElement
from liquid.expression import Identifier
from liquid.expression import Boolean
from liquid.expression import StringLiteral
from liquid.expression import IntegerLiteral
from liquid.expression import FloatLiteral
from liquid.expression import RangeLiteral
from liquid.expression import InfixExpression

from liquid.expressions import parse_boolean_expression
from liquid.expressions import parse_boolean_expression_with_parens
from liquid.expressions.boolean.parse import PrefixExpression


class Case(NamedTuple):
    description: str
    expression: str
    expect: Union[Expression, str]


class ParseBooleanExpressionTestCase(unittest.TestCase):
    """Test cases for parsing standard boolean expressions."""

    test_cases = [
        Case(
            "string literal double quotes",
            '"foobar"',
            BooleanExpression(
                expression=StringLiteral("foobar"),
            ),
        ),
        Case(
            "integer literal",
            "7",
            BooleanExpression(
                expression=IntegerLiteral(7),
            ),
        ),
        Case(
            "negative integer literal statement expression",
            "-7",
            BooleanExpression(
                expression=IntegerLiteral(-7),
            ),
        ),
        Case(
            "float literal statement expression",
            "3.14",
            BooleanExpression(
                expression=FloatLiteral(3.14),
            ),
        ),
        Case(
            "negative float literal statement expression",
            "-3.14",
            BooleanExpression(expression=FloatLiteral(-3.14)),
        ),
        Case(
            "single identifier statement expression",
            "collection",
            BooleanExpression(
                expression=Identifier(
                    path=[IdentifierPathElement("collection")],
                ),
            ),
        ),
        Case(
            "chained identifier",
            "collection.products",
            BooleanExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                    ],
                ),
            ),
        ),
        Case(
            "keyword true",
            "true",
            BooleanExpression(
                expression=Boolean(True),
            ),
        ),
        Case(
            "keyword false",
            "false",
            BooleanExpression(
                expression=Boolean(False),
            ),
        ),
        Case(
            "boolean equality",
            "true == true",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="==",
                    right=Boolean(True),
                ),
            ),
        ),
        Case(
            "boolean inequality",
            "true != false",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="!=",
                    right=Boolean(False),
                ),
            ),
        ),
        Case(
            "boolean inequality alternate",
            "true <> false",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="<>",
                    right=Boolean(False),
                ),
            ),
        ),
        Case(
            "identifier greater than literal",
            "user.age > 21",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[
                            IdentifierPathElement("user"),
                            IdentifierPathElement("age"),
                        ],
                    ),
                    operator=">",
                    right=IntegerLiteral(21),
                ),
            ),
        ),
        Case(
            "identifier less than literal",
            "age < 18",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[IdentifierPathElement("age")],
                    ),
                    operator="<",
                    right=IntegerLiteral(18),
                ),
            ),
        ),
        Case(
            "identifier less than or equal to literal",
            "age <= 18",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[IdentifierPathElement("age")],
                    ),
                    operator="<=",
                    right=IntegerLiteral(18),
                ),
            ),
        ),
        Case(
            "identifier greater than or equal to literal",
            "age >= 18",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[IdentifierPathElement("age")],
                    ),
                    operator=">=",
                    right=IntegerLiteral(18),
                ),
            ),
        ),
        Case(
            "boolean or boolean",
            "true or false",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="or",
                    right=Boolean(False),
                ),
            ),
        ),
        Case(
            "identifier contains string",
            "product.tags contains 'sale'",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[
                            IdentifierPathElement("product"),
                            IdentifierPathElement("tags"),
                        ],
                    ),
                    operator="contains",
                    right=StringLiteral("sale"),
                ),
            ),
        ),
        Case(
            "identifier equals a range expression",
            "foo == (1..3)",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[
                            IdentifierPathElement("foo"),
                        ],
                    ),
                    operator="==",
                    right=RangeLiteral(
                        IntegerLiteral(1),
                        IntegerLiteral(3),
                    ),
                ),
            ),
        ),
        Case(
            "chained identifier with bracketed index and string",
            "users[0]['age'] > 21",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[
                            IdentifierPathElement("users"),
                            IdentifierPathElement(0),
                            IdentifierPathElement("age"),
                        ],
                    ),
                    operator=">",
                    right=IntegerLiteral(21),
                ),
            ),
        ),
        Case(
            "multiple lines",
            "users[0]['age'] \n>  \n21",
            BooleanExpression(
                expression=InfixExpression(
                    left=Identifier(
                        path=[
                            IdentifierPathElement("users"),
                            IdentifierPathElement(0),
                            IdentifierPathElement("age"),
                        ],
                    ),
                    operator=">",
                    right=IntegerLiteral(21),
                ),
            ),
        ),
    ]

    def test_parse_boolean_expression(self):
        """Test that we can parse standard boolean expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_boolean_expression(case.expression)
                self.assertEqual(expr, case.expect)

    def test_parse_boolean_expression_with_parens(self):
        """Test that the non-standard boolean expression parser is backwards
        compatible with standard boolean expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_boolean_expression_with_parens(case.expression)
                self.assertEqual(expr, case.expect)

    def test_parse_boolean_expression_precedence(self):
        """Test that we get the expected precedence when parsing boolean expressions."""
        test_cases = [
            Case(
                "boolean and logic operators",
                "true and false and false or true",
                "(True and (False and (False or True)))",
            ),
            Case(
                "equality and logic operators",
                "user.name == 'bob' and user.age < 50 or collection.title == 'offers'",
                (
                    "((user.name == 'bob') "
                    "and ((user.age < 50) "
                    "or (collection.title == 'offers')"
                    "))"
                ),
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                expr = parse_boolean_expression(case.expression)
                self.assertEqual(str(expr), case.expect)

            with self.subTest(msg="[with parens] " + case.description):
                expr = parse_boolean_expression_with_parens(case.expression)
                self.assertEqual(str(expr), case.expect)


class ParseBooleanNotExpressionTestCase(unittest.TestCase):
    """Test cases for parsing non-standard boolean expressions."""

    test_cases = [
        Case(
            "not true",
            "not true",
            BooleanExpression(
                expression=PrefixExpression(
                    "not",
                    right=Boolean(True),
                ),
            ),
        ),
        Case(
            "right associative",
            "true and false and false or true",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="and",
                    right=InfixExpression(
                        left=Boolean(False),
                        operator="and",
                        right=InfixExpression(
                            left=Boolean(False),
                            operator="or",
                            right=Boolean(True),
                        ),
                    ),
                ),
            ),
        ),
        Case(
            "grouped boolean",
            "(true and false and false) or true",
            BooleanExpression(
                expression=InfixExpression(
                    left=InfixExpression(
                        left=Boolean(True),
                        operator="and",
                        right=InfixExpression(
                            left=Boolean(False),
                            operator="and",
                            right=Boolean(False),
                        ),
                    ),
                    operator="or",
                    right=Boolean(True),
                ),
            ),
        ),
        Case(
            "parens",
            "(true and false and false)",
            BooleanExpression(
                expression=InfixExpression(
                    left=Boolean(True),
                    operator="and",
                    right=InfixExpression(
                        left=Boolean(False),
                        operator="and",
                        right=Boolean(False),
                    ),
                ),
            ),
        ),
    ]

    def test_parser_with_parens(self) -> None:
        """Test that we can parse boolean expressions that support logical `not`
        and grouping terms with parentheses."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_boolean_expression_with_parens(case.expression)
                self.assertEqual(expr, case.expect)



"""Node and tag definitions for `macro` and `call`."""
from __future__ import annotations

import itertools
import sys

from typing import Optional
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import TextIO
from typing import Union
from typing import TYPE_CHECKING

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import is_undefined
from liquid.context import Undefined
from liquid.context import Context
from liquid.context import ReadOnlyChainMap

from liquid.expression import Expression
from liquid.expression import NIL

from liquid.expressions import parse_call_arguments
from liquid.expressions import parse_macro_arguments
from liquid.expressions.arguments.parse import Argument

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.builtin.tags.include_tag import TAG_INCLUDE

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment


TAG_MACRO = sys.intern("macro")
TAG_ENDMACRO = sys.intern("endmacro")
TAG_CALL = sys.intern("call")


class Macro(NamedTuple):
    """A macro block and its arguments."""

    args: List[Argument]
    block: BlockNode


class MacroNode(Node):
    """Parse tree node representing a macro."""

    __slots__ = ("tok", "name", "args", "block")

    def __init__(
        self,
        tok: Token,
        name: str,
        args: List[Argument],
        block: BlockNode,
    ):
        self.tok = tok
        self.name = name
        self.args = args
        self.block = block

    def __str__(self) -> str:  # pragma: no cover
        args: List[str] = []
        for arg, default in self.args:
            if default is not NIL:
                args.append(f"{arg}={default}")
            else:
                args.append(arg)
        return f"def {self.name}({', '.join(args)}) {{ {self.block} }}"

    def __repr__(self) -> str:  # pragma: no cover
        return f"MacroNode(tok={self.tok}, name={self.name}, block='{self.block}')"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        if "macros" not in context.tag_namespace:
            context.tag_namespace["macros"] = {}

        context.tag_namespace["macros"][self.name] = Macro(self.args, self.block)
        return False

    def children(self) -> List[ChildNode]:
        _children = [
            ChildNode(
                linenum=self.tok.linenum,
                node=self.block,
                block_scope=[key for key, _ in self.args],
            )
        ]
        for _, expr in self.args:
            if expr != NIL:
                _children.append(ChildNode(linenum=self.tok.linenum, expression=expr))
        return _children


class CallNode(Node):
    """Parse tree node representing a call to a macro."""

    __slots__ = ("tok", "name", "args", "kwargs")

    def __init__(
        self,
        tok: Token,
        name: str,
        args: List[Expression],
        kwargs: List[Argument],
    ):
        self.tok = tok
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:  # pragma: no cover
        args = [str(expr) for expr in self.args]
        for name, expr in self.kwargs:
            args.append(f"{name}: {expr}")
        return f"{self.name}({', '.join(args)})"

    def __repr__(self) -> str:  # pragma: no cover
        return f"CallNode(tok={self.tok}, name={self.name})"

    def _make_context(self, context: Context, macro: Macro) -> Context:
        args = dict(macro.args)
        macro_names = [key for key, val in macro.args]

        # Bind positional arguments to names. If there are more positional arguments
        # than names defined in the macro, they'll be stored in `excess_args`.
        excess_args: List[Expression] = []
        for name, expr in itertools.zip_longest(macro_names, self.args, fillvalue=NIL):
            if name == NIL:
                excess_args.append(expr)
                continue
            if expr == NIL:
                break
            assert isinstance(name, str)
            args[name] = expr

        # NOTE: This has the potential to override a positional argument with a keyword.
        # We also silently take the last keyword argument if a keyword is given more
        # than once.

        # Update default and/or missing arguments with keyword arguments.
        excess_kwargs: Dict[str, Expression] = {}
        for name, expr in self.kwargs:
            assert isinstance(name, str)
            if name in macro_names:
                args[name] = expr
            else:
                excess_kwargs[name] = expr

        excess: Dict[str, object] = {
            "kwargs": {
                name: expr.evaluate(context) for name, expr in excess_kwargs.items()
            },
            "args": [expr.evaluate(context) for expr in excess_args],
        }

        # NOTE: default arguments are bound late.
        bound_args: Dict[str, object] = {}
        for name, expr in args.items():
            if expr == NIL:
                bound_args[name] = context.env.undefined(name)
            else:
                assert isinstance(expr, Expression)
                assert isinstance(name, str)
                bound_args[name] = expr.evaluate(context)

        namespace = ReadOnlyChainMap(bound_args, excess)
        return context.copy(namespace, disabled_tags=[TAG_INCLUDE])

    def _get_macro(self, context: Context) -> Union[Macro, Undefined]:
        macro = context.tag_namespace.get("macros", {}).get(
            self.name, context.env.undefined(self.name)
        )
        assert isinstance(macro, (Macro, Undefined))
        return macro

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        macro = self._get_macro(context)

        if is_undefined(macro):
            buffer.write(str(macro))
            return False

        assert isinstance(macro, Macro)

        ctx = self._make_context(context, macro)
        macro.block.render(ctx, buffer)

        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        macro = self._get_macro(context)

        if is_undefined(macro):
            buffer.write(str(macro))
            return False

        assert isinstance(macro, Macro)

        ctx = self._make_context(context, macro)
        await macro.block.render_async(ctx, buffer)

        return True

    def children(self) -> List[ChildNode]:
        return [
            *[
                ChildNode(linenum=self.tok.linenum, expression=expr)
                for expr in self.args
            ],
            *[
                ChildNode(linenum=self.tok.linenum, expression=val)
                for _, val in self.kwargs
            ],
        ]


class MacroTag(Tag):
    """Macro tag definition."""

    name = TAG_MACRO
    end = TAG_ENDMACRO

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_MACRO)
        tok = stream.current
        stream.next_token()

        # Parse the expression
        expect(stream, TOKEN_EXPRESSION)
        name, args = parse_macro_arguments(stream.current.value, linenum=tok.linenum)
        stream.next_token()

        # Parse the block
        block = self.parser.parse_block(stream, (TAG_ENDMACRO, TOKEN_EOF))
        expect(stream, TOKEN_TAG, value=TAG_ENDMACRO)

        return MacroNode(tok=tok, name=name, args=args, block=block)


class CallTag(Tag):
    """Call tag definition."""

    name = TAG_CALL
    block = False

    def parse(self, stream: TokenStream) -> CallNode:
        expect(stream, TOKEN_TAG, value=TAG_CALL)
        tok = stream.current

        stream.next_token()
        expect(stream, TOKEN_EXPRESSION)
        name, _args = parse_call_arguments(stream.current.value, linenum=tok.linenum)

        # Args can be positional (no default), or keyword (with default).
        args: List[Expression] = []
        kwargs: List[Argument] = []

        for key, val in _args:
            if key is None:
                args.append(val)
            else:
                kwargs.append((key, val))

        return CallNode(tok=tok, name=name, args=args, kwargs=kwargs)



"""Test the "drop" API."""
# pylint: disable=missing-class-docstring
import unittest

from typing import Generic
from typing import NamedTuple
from typing import TypeVar

from liquid import Environment


class Case(NamedTuple):
    description: str
    template: str
    context: dict
    expect: str


DropT = TypeVar("DropT")


class Drop(Generic[DropT]):
    """Mock generic drop that wraps a primitive value."""

    def __init__(self, val: DropT):
        self.val = val

    def __liquid__(self) -> DropT:
        return self.val

    def __str__(self) -> str:
        return str(self.val)


class IntDrop:
    """Mock drop that wraps an integer value."""

    def __init__(self, val: int):
        self.val = val

    def __int__(self) -> int:
        return self.val

    def __str__(self) -> str:
        return "one"

    def __liquid__(self) -> int:
        return int(self)


class ConfusedDrop:
    """Mock drop that will evaluate to a boolean when used as an array index or hash
    key, or when used in a conditional expression, but will render as a string."""

    def __init__(self, val: bool):
        self.val = val

    def __liquid__(self) -> bool:
        return self.val

    def __str__(self) -> str:
        return "yay" if self.val else "nay"


class DropAPITestCase(unittest.TestCase):
    """Test the drop API."""

    def test_to_liquid(self):
        """Test that user defined classes can be used as primitives."""
        test_cases = [
            Case(
                description="int drop",
                template=r"{{ drop }}",
                context={"drop": IntDrop(1)},
                expect="one",
            ),
            Case(
                description="int drop as array index",
                template=r"{{ foo[drop] }}",
                context={"foo": ["a", "b", "c"], "drop": IntDrop(1)},
                expect="b",
            ),
            Case(
                description="int drop as hash key",
                template=r"{{ foo[drop] }}",
                context={"foo": {1: "a", 2: "b"}, "drop": IntDrop(1)},
                expect="a",
            ),
            Case(
                description="int drop as filter left value",
                template=r"{{ drop | plus: 1 }}",
                context={"drop": IntDrop(1)},
                expect="1",
            ),
            Case(
                description="int drop in boolean expression",
                template=r"{% if drop == 1 %}one{% endif %}",
                context={"drop": IntDrop(1)},
                expect="one",
            ),
            Case(
                description="int drop is less than int drop",
                template=r"{% if some < other %}hello{% endif %}",
                context={"some": IntDrop(1), "other": IntDrop(2)},
                expect="hello",
            ),
            Case(
                description="int drop in case expression",
                template=r"{% case drop %}{% when 1 %}one{% endcase %}",
                context={"drop": IntDrop(1)},
                expect="one",
            ),
            Case(
                description="boolean drop",
                template=r"{{ drop }}",
                context={"drop": Drop[bool](False)},
                expect="False",
            ),
            Case(
                description="bool drop in boolean expression",
                template=r"{% if drop == true %}one{% endif %}",
                context={"drop": Drop[bool](True)},
                expect="one",
            ),
            Case(
                description="false bool drop in boolean expression",
                template=r"{% if drop == true %}one{% endif %}",
                context={"drop": Drop[bool](False)},
                expect="",
            ),
            Case(
                description="false bool drop is False",
                template=r"{% if drop %}one{% endif %}",
                context={"drop": Drop[bool](False)},
                expect="",
            ),
            Case(
                description="bool drop in unless expression",
                template=r"{% unless drop %}hello{% endunless %}",
                context={"drop": Drop[bool](True)},
                expect="",
            ),
            Case(
                description="confused drop",
                template=r"{{ drop }}",
                context={"drop": ConfusedDrop(False)},
                expect="nay",
            ),
            Case(
                description="filtered confused drop",
                template=r"{{ drop | upcase }}",
                context={"drop": ConfusedDrop(False)},
                expect="NAY",
            ),
            Case(
                description="compare confused drop",
                template=r"{% if drop %}{{ drop | upcase }}{% endif %}",
                context={"drop": ConfusedDrop(False)},
                expect="",
            ),
        ]

        env = Environment()

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)



"""Tag and node definition for the mock "form" tag."""

import sys

from collections import abc

from typing import Optional
from typing import TextIO

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_EOF

from liquid.tag import Tag
from liquid import ast
from liquid.stream import TokenStream
from liquid.lex import tokenize_identifier
from liquid.expression import Identifier
from liquid.context import Context

from liquid.parse import get_parser
from liquid.parse import expect
from liquid.parse import parse_identifier


TAG_FORM = sys.intern("form")
TAG_ENDFORM = sys.intern("endform")

END_FORMBLOCK = frozenset((TAG_ENDFORM,))


class CommentFormNode(ast.Node):
    """Parse tree node for the mock "form" tag."""

    __slots__ = ("tok", "article", "block")

    def __init__(self, tok: Token, article: Identifier, block: ast.BlockNode):
        self.tok = tok
        self.article = article
        self.block = block

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        article = self.article.evaluate(context)
        assert isinstance(article, abc.Mapping)

        form = {
            "posted_successfully?": context.get("posted_successfully", True),
            "errors": context.get(["comment", "errors"], []),
            "author": context.get(["comment", "author"]),
            "email": context.get(["comment", "email"]),
            "body": context.get(["comment", "body"]),
        }

        buffer.write(
            f'<form id="article-{article["id"]}-comment-form" '
            'class="comment-form" method="post" action="">\n'
        )

        with context.extend({"form": form}):
            self.block.render(context, buffer)

        buffer.write("\n</form>")


class CommentFormTag(Tag):
    """The mock "paginate" tag."""

    name = TAG_FORM
    end = TAG_ENDFORM

    def parse(self, stream: TokenStream) -> ast.Node:
        parser = get_parser(self.env)

        expect(stream, TOKEN_TAG, value=TAG_FORM)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_identifier(stream.current.value))

        expect(expr_stream, TOKEN_IDENTIFIER)
        article = parse_identifier(expr_stream)
        expr_stream.next_token()

        # End of expression
        expect(expr_stream, TOKEN_EOF)

        # Advance the stream passed the expression and read the block.
        stream.next_token()
        block = parser.parse_block(stream, end=END_FORMBLOCK)
        expect(stream, TOKEN_TAG, value=TAG_ENDFORM)

        return CommentFormNode(tok, article=article, block=block)



"""Liquid render tests."""
# pylint: disable=missing-function-docstring,too-many-public-methods,missing-class-docstring
import asyncio
import unittest

from typing import List

from liquid.environment import Environment
from liquid.template import AwareBoundTemplate
from liquid.mode import Mode
from liquid.loaders import DictLoader

from liquid import golden
from liquid.exceptions import Error
from liquid.golden.case import Case


class RenderTestCase(unittest.TestCase):
    def _test(
        self,
        test_cases: List[Case],
        template_class=AwareBoundTemplate,
        tolerance=Mode.STRICT,
    ):
        """Run all tests in `test_cases` in sync and async modes."""
        self._test_sync(test_cases, template_class, tolerance)
        self._test_async(test_cases, template_class, tolerance)

    def _test_sync(
        self,
        test_cases: List[Case],
        template_class=AwareBoundTemplate,
        tolerance=Mode.STRICT,
    ):
        """Helper method for testing lists of test cases."""
        for case in test_cases:
            env = Environment(
                loader=DictLoader(case.partials),
                tolerance=tolerance,
            )
            env.template_class = template_class

            with self.subTest(msg=case.description, mode=tolerance):
                if case.error:
                    with self.assertRaises(Error):
                        template = env.from_string(case.template, globals=case.globals)
                        result = template.render()
                else:
                    template = env.from_string(case.template, globals=case.globals)
                    result = template.render()
                    self.assertEqual(result, case.expect)

    def _test_async(
        self,
        test_cases: List[Case],
        template_class=AwareBoundTemplate,
        tolerance=Mode.STRICT,
    ):
        """Helper method for table driven testing of asynchronous rendering."""

        async def coro(template):
            return await template.render_async()

        for case in test_cases:
            env = Environment(
                loader=DictLoader(case.partials),
                tolerance=tolerance,
            )
            env.template_class = template_class

            with self.subTest(msg=case.description, asynchronous=True, mode=tolerance):
                if case.error:
                    with self.assertRaises(Error):
                        template = env.from_string(case.template, globals=case.globals)
                        result = asyncio.run(coro(template))
                else:
                    template = env.from_string(case.template, globals=case.globals)
                    result = asyncio.run(coro(template))
                    self.assertEqual(result, case.expect)


class RenderTagsTestCase(RenderTestCase):
    def test_literal(self):
        """Test that we can render template literals."""
        self._test(golden.not_liquid.cases)

    def test_whitespace_control(self):
        """Test that we can control whitespace."""
        self._test(golden.whitespace_control.cases)

    def test_output_statement(self):
        """Test that we can render output statements."""
        self._test(golden.output_statement.cases)

    def test_echo_tag(self):
        """Test that we can render `echo` tags."""
        self._test(golden.echo_tag.cases)

    def test_assign_tag(self):
        """Test that we can render assigned variables."""
        self._test(golden.assign_tag.cases)

    def test_if_tag(self):
        """Test that we can render `if` tags."""
        self._test(golden.if_tag.cases)

    def test_comment_tag(self):
        """Test that we can render comment tags."""
        self._test(golden.comment_tag.cases)

    def test_unless_tag(self):
        """Test that we can render `unless` tags."""
        self._test(golden.unless_tag.cases)

    def test_capture_tag(self):
        """Test that we can render `capture` tags."""
        self._test(golden.capture_tag.cases)

    def test_case_tag(self):
        """Test that we can render `case` tags."""
        self._test(golden.case_tag.cases)

    def test_cycle_tag(self):
        """Test that we can render `cycle` tags."""
        self._test(golden.cycle_tag.cases)

    def test_decrement_tag(self):
        """Test that we can render `decrement` tags."""
        self._test(golden.decrement_tag.cases)

    def test_increment_tag(self):
        """Test that we can render `increment` tags."""
        self._test(golden.increment_tag.cases)

    def test_for_tag(self):
        """Test that we can render `for` tags."""
        self._test(golden.for_tag.cases)

    def test_raw_tag(self):
        """Test that we can render `raw` tags."""
        self._test(golden.raw_tag.cases)

    def test_tablerow_tag(self):
        """Test that we can render `tablerow` tags."""
        self._test(golden.tablerow_tag.cases)

    def test_liquid_tag(self):
        """Test that we can render `liquid` tags."""
        self._test(golden.liquid_tag.cases)

    def test_illegal(self):
        """Test that we can render an `illegal` token in LAX mode."""
        self._test(golden.illegal.cases)

    def test_include_tag(self):
        """Test that we can render `include` tags."""
        self._test(golden.include_tag.cases)

    def test_render_tag(self):
        """Test that we can render `render` tags."""
        self._test(golden.render_tag.cases)

    def test_ifchanged_tag(self):
        """Test that we can render `ifchanged` tags."""
        self._test(golden.ifchanged_tag.cases)

    def test_inline_comment_tag(self):
        """Test that we can render `#` tags."""
        self._test(golden.inline_comment_tag.cases)


class RenderFiltersTestCase(RenderTestCase):
    def test_render_concat_filter(self):
        """Test that we render the `concat` filter correctly."""
        self._test(golden.concat_filter.cases)

    def test_render_first_filter(self):
        """Test that we render the `first` filter correctly."""
        self._test(golden.first_filter.cases)

    def test_render_join_filter(self):
        """Test that we render the `join` filter correctly."""
        self._test(golden.join_filter.cases)

    def test_render_last_filter(self):
        """Test that we render the `last` filter correctly."""
        self._test(golden.last_filter.cases)

    def test_render_map_filter(self):
        """Test that we render the `map` filter correctly."""
        self._test(golden.map_filter.cases)

    def test_render_reverse_filter(self):
        """Test that we render the `reverse` filter correctly."""
        self._test(golden.reverse_filter.cases)

    def test_render_sort_filter(self):
        """Test that we render the `sort` filter correctly."""
        self._test(golden.sort_filter.cases)

    def test_render_sort_natural_filter(self):
        """Test that we render the `sort_natural` filter correctly."""
        self._test(golden.sort_natural_filter.cases)

    def test_render_where_filter(self):
        """Test that we render the `where` filter correctly."""
        self._test(golden.where_filter.cases)

    def test_render_uniq_filter(self):
        """Test that we render the `uniq` filter correctly."""
        self._test(golden.uniq_filter.cases)

    def test_render_compact_filter(self):
        """Test that we render the `compact` filter correctly."""
        self._test(golden.compact_filter.cases)

    def test_render_abs_filter(self):
        """Test that we render the `abs` filter correctly."""
        self._test(golden.abs_filter.cases)

    def test_render_at_most_filter(self):
        """Test that we render the `at_most` filter correctly."""
        self._test(golden.at_most_filter.cases)

    def test_render_at_least_filter(self):
        """Test that we render the `at_least` filter correctly."""
        self._test(golden.at_least_filter.cases)

    def test_render_ceil_filter(self):
        """Test that we render the `ceil` filter correctly."""
        self._test(golden.ceil_filter.cases)

    def test_render_floor_filter(self):
        """Test that we render the `floor` filter correctly."""
        self._test(golden.floor_filter.cases)

    def test_render_divided_by_filter(self):
        """Test that we render the `divided_by` filter correctly."""
        self._test(golden.divided_by_filter.cases)

    def test_render_minus_filter(self):
        """Test that we render the `minus` filter correctly."""
        self._test(golden.minus_filter.cases)

    def test_render_plus_filter(self):
        """Test that we render the `plus` filter correctly."""
        self._test(golden.plus_filter.cases)

    def test_render_round_filter(self):
        """Test that we render the `round` filter correctly."""
        self._test(golden.round_filter.cases)

    def test_render_times_filter(self):
        """Test that we render the `times` filter correctly."""
        self._test(golden.times_filter.cases)

    def test_render_modulo_filter(self):
        """Test that we render the `modulo` filter correctly."""
        self._test(golden.modulo_filter.cases)

    def test_render_size_filter(self):
        """Test that we render the `size` filter correctly."""
        self._test(golden.size_filter.cases)

    def test_render_default_filter(self):
        """Test that we render the `default` filter correctly."""
        self._test(golden.default_filter.cases)

    def test_render_date_filter(self):
        """Test that we render the `date` filter correctly."""
        self._test(golden.date_filter.cases)

    def test_render_capitalize_filter(self):
        """Test that we can render the `capitalize` filter correctly."""
        self._test(golden.capitalize_filter.cases)

    def test_render_append_filter(self):
        """Test that we can render the `append` filter correctly."""
        self._test(golden.append_filter.cases)

    def test_render_downcase_filter(self):
        """Test that we can render the `downcase` filter correctly."""
        self._test(golden.downcase_filter.cases)

    def test_render_escape_filter(self):
        """Test that we can render the `escape` filter correctly."""
        self._test(golden.escape_filter.cases)

    def test_render_escape_once_filter(self):
        """Test that we can render the `escape_once` filter correctly."""
        self._test(golden.escape_once_filter.cases)

    def test_render_lstrip_filter(self):
        """Test that we can render the `lstrip` filter correctly."""
        self._test(golden.lstrip_filter.cases)

    def test_render_newline_to_br_filter(self):
        """Test that we can render the `newline_to_br` filter correctly."""
        self._test(golden.newline_to_br_filter.cases)

    def test_render_prepend_filter(self):
        """Test that we can render the `prepend` filter correctly."""
        self._test(golden.prepend_filter.cases)

    def test_render_remove_filter(self):
        """Test that we can render the `remove` filter correctly."""
        self._test(golden.remove_filter.cases)

    def test_render_remove_first_filter(self):
        """Test that we can render the `remove_first` filter correctly."""
        self._test(golden.remove_first_filter.cases)

    def test_render_remove_last_filter(self):
        """Test that we can render the `remove_last` filter correctly."""
        self._test(golden.remove_last_filter.cases)

    def test_render_replace_filter(self):
        """Test that we can render the `replace` filter correctly."""
        self._test(golden.replace_filter.cases)

    def test_render_replace_first_filter(self):
        """Test that we can render the `replace_first` filter correctly."""
        self._test(golden.replace_first_filter.cases)

    def test_render_replace_last_filter(self):
        """Test that we can render the `replace_last` filter correctly."""
        self._test(golden.replace_last_filter.cases)

    def test_render_slice_filter(self):
        """Test that we can render the `slice` filter correctly."""
        self._test(golden.slice_filter.cases)

    def test_render_split_filter(self):
        """Test that we can render the `split` filter correctly."""
        self._test(golden.split_filter.cases)

    def test_render_upcase_filter(self):
        """Test that we can render the `upcase` filter correctly."""
        self._test(golden.upcase_filter.cases)

    def test_render_strip_filter(self):
        """Test that we can render the `strip` filter correctly."""
        self._test(golden.strip_filter.cases)

    def test_render_rstrip_filter(self):
        """Test that we can render the `rstrip` filter correctly."""
        self._test(golden.rstrip_filter.cases)

    def test_render_strip_html_filter(self):
        """Test that we can render the `strip_html` filter correctly."""
        self._test(golden.strip_html_filter.cases)

    def test_render_strip_newlines_filter(self):
        """Test that we can render the `strip_newlines` filter correctly."""
        self._test(golden.strip_newlines_filter.cases)

    def test_render_truncate_filter(self):
        """Test that we can render the `truncate` filter correctly."""
        self._test(golden.truncate_filter.cases)

    def test_render_truncatewords_filter(self):
        """Test that we can render the `truncatewords` filter correctly."""
        self._test(golden.truncatewords_filter.cases)

    def test_render_url_encode_filter(self):
        """Test that we can render the `url_encode` filter correctly."""
        self._test(golden.url_encode_filter.cases)

    def test_render_url_decode_filter(self):
        """Test that we can render the `url_decode` filter correctly."""
        self._test(golden.url_decode_filter.cases)

    def test_render_base64_encode_filter(self):
        """Test that we can render the `base64_encode` filter correctly."""
        self._test(golden.base64_encode_filter.cases)

    def test_render_base64_decode_filter(self):
        """Test that we can render the `base64_decode` filter correctly."""
        self._test(golden.base64_decode_filter.cases)

    def test_render_base64_url_safe_encode_filter(self):
        """Test that we can render the `base64_url_safe_encode` filter correctly."""
        self._test(golden.base64_url_safe_encode_filter.cases)

    def test_render_base64_url_safe_decode_filter(self):
        """Test that we can render the `base64_url_safe_decode` filter correctly."""
        self._test(golden.base64_url_safe_decode_filter.cases)


class RenderMiscTestCase(RenderTestCase):
    def test_render_range_objects(self):
        """Test that we can render range objects."""
        self._test(golden.range_objects.cases)

    def test_special_properties(self):
        """Test that we can use special, built-in properties."""
        self._test(golden.special.cases)

    def test_identifiers(self):
        """Test permitted identifiers."""
        self._test(golden.identifiers.cases)



"""Test math filter functions."""
# pylint: disable=too-many-public-methods,too-many-lines,missing-class-docstring
import unittest

from functools import partial
from inspect import isclass

from typing import NamedTuple
from typing import Any
from typing import List
from typing import Dict

from liquid.environment import Environment

from liquid.exceptions import FilterArgumentError

from liquid.builtin.filters.math import abs_
from liquid.builtin.filters.math import at_most
from liquid.builtin.filters.math import at_least
from liquid.builtin.filters.math import ceil
from liquid.builtin.filters.math import divided_by
from liquid.builtin.filters.math import floor
from liquid.builtin.filters.math import minus
from liquid.builtin.filters.math import plus
from liquid.builtin.filters.math import round_
from liquid.builtin.filters.math import times
from liquid.builtin.filters.math import modulo


class Case(NamedTuple):
    """Table driven test case helper."""

    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class MathFilterTestCase(unittest.TestCase):
    """Test math filter functions."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases):
        if getattr(func, "with_environment", False):
            func = partial(func, environment=self.env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, FilterArgumentError
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_abs(self):
        """Test abs filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string integer",
                val="5",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative string integer",
                val="-5",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=5.1,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=5.1,
            ),
            Case(
                description="unexpected argument",
                val=-3,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=0,
            ),
        ]

        self._test(abs_, test_cases)

    def test_at_most(self):
        """Test at_most filter function."""

        test_cases = [
            Case(
                description="positive integer < arg",
                val=5,
                args=[8],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive integer > arg",
                val=8,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer < arg",
                val=-8,
                args=[5],
                kwargs={},
                expect=-8,
            ),
            Case(
                description="positive integer == arg",
                val=5,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float < arg",
                val=5.4,
                args=[8.9],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="positive float > arg",
                val=8.4,
                args=[5.9],
                kwargs={},
                expect=5.9,
            ),
            Case(
                description="positive string > arg",
                val="9",
                args=["8"],
                kwargs={},
                expect=8,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[5],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined argument",
                val=5,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=0,
            ),
            Case(
                description="unacceptable left value",
                val="abc",
                args=[-2],
                kwargs={},
                expect=-2,
            ),
        ]

        self._test(at_most, test_cases)

    def test_at_least(self):
        """Test at_least filter function."""

        test_cases = [
            Case(
                description="positive integer < arg",
                val=5,
                args=[8],
                kwargs={},
                expect=8,
            ),
            Case(
                description="positive integer > arg",
                val=8,
                args=[5],
                kwargs={},
                expect=8,
            ),
            Case(
                description="negative integer < arg",
                val=-8,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive integer == arg",
                val=5,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float < arg",
                val=5.4,
                args=[8.9],
                kwargs={},
                expect=8.9,
            ),
            Case(
                description="positive float > arg",
                val=8.4,
                args=[5.9],
                kwargs={},
                expect=8.4,
            ),
            Case(
                description="positive string > arg",
                val="9",
                args=["8"],
                kwargs={},
                expect=9,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="undefined argument",
                val=5,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=5,
            ),
            Case(
                description="unacceptable left value",
                val="abc",
                args=[2],
                kwargs={},
                expect=2,
            ),
        ]

        self._test(at_least, test_cases)

    def test_ceil(self):
        """Test ceil filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="unexpected argument",
                val=-3.1,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=0,
            ),
        ]

        self._test(ceil, test_cases)

    def test_floor(self):
        """Test floor filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=-6,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=-6,
            ),
            Case(
                description="unexpected argument",
                val=-3.1,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect=0,
            ),
        ]

        self._test(floor, test_cases)

    def test_divided_by(self):
        """Test divided_by filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=5,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=5.0,
            ),
            Case(
                description="integer division",
                val=9,
                args=[2],
                kwargs={},
                expect=4,
            ),
            Case(
                description="float division",
                val=20,
                args=[7.0],
                kwargs={},
                expect=2.857142857142857,
            ),
            Case(
                description="string value and argument",
                val="10",
                args=["2.0"],
                kwargs={},
                expect=5.0,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=0,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined argument",
                val=10,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(divided_by, test_cases)

    def test_minus(self):
        """Test minus filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=8,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=8.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[2.2],
                kwargs={},
                expect=7.9,
            ),
            Case(
                description="string value and string arg",
                val="10.1",
                args=["2.2"],
                kwargs={},
                expect=7.9,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=-2.0,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=10,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=-1,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=-2,
            ),
            Case(
                description="undefined argument",
                val=10,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=10,
            ),
        ]

        self._test(minus, test_cases)

    def test_plus(self):
        """Test plus filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=12,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=12.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[2.2],
                kwargs={},
                expect=12.3,
            ),
            Case(
                description="string value and string arg",
                val="10.1",
                args=["2.2"],
                kwargs={},
                expect=12.3,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=2.0,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=10,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=1,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=2,
            ),
            Case(
                description="undefined argument",
                val=10,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=10,
            ),
        ]

        self._test(plus, test_cases)

    def test_round(self):
        """Test round filter function."""

        test_cases = [
            Case(
                description="float round down",
                val=5.1,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="float round up",
                val=5.6,
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="float as a string",
                val="5.6",
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="string argument",
                val=5.666,
                args=["1"],
                kwargs={},
                expect=5.7,
            ),
            Case(
                description="decimal places",
                val="5.666666",
                args=[2],
                kwargs={},
                expect=5.67,
            ),
            Case(
                description="integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined argument",
                val=5.666,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=6,
            ),
        ]

        self._test(round_, test_cases)

    def test_times(self):
        """Test times filter function."""

        test_cases = [
            Case(
                description="int times int",
                val=5,
                args=[2],
                kwargs={},
                expect=10,
            ),
            Case(
                description="int times float",
                val=5,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="int times float",
                val=5,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="float times float",
                val=5.0,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="string times string",
                val="5.0",
                args=["2.1"],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="negative multiplication",
                val=-5,
                args=[2],
                kwargs={},
                expect=-10,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined argument",
                val=5,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=0,
            ),
        ]

        self._test(times, test_cases)

    def test_modulo(self):
        """Test modulo filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=0.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[7.0],
                kwargs={},
                expect=3.1,
            ),
            Case(
                description="string value and argument",
                val="10",
                args=["2.0"],
                kwargs={},
                expect=0.0,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=0,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="undefined argument",
                val=5,
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(modulo, test_cases)



"""Functions for parsing loop expressions. Like those found in `for` and `tablerow`
tags."""
from typing import Callable
from typing import Dict
from typing import Tuple

from liquid.limits import to_int

from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_IN
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_LIMIT
from liquid.token import TOKEN_OFFSET
from liquid.token import TOKEN_COLS
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_CONTINUE
from liquid.token import TOKEN_REVERSED
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_COMMA

from liquid.expression import Continue
from liquid.expression import CONTINUE
from liquid.expression import IntegerLiteral
from liquid.expression import LoopExpression
from liquid.expression import LoopIterable
from liquid.expression import LoopArgument

from liquid.expressions.common import parse_identifier
from liquid.expressions.common import parse_integer_literal
from liquid.expressions.common import parse_float_literal
from liquid.expressions.common import make_parse_range

from liquid.expressions.filtered.parse import parse_obj as parse_simple_obj
from liquid.expressions.stream import TokenStream
from liquid.expressions.loop.lex import tokenize

from liquid.exceptions import LiquidSyntaxError

LOOP_ARGS = frozenset(
    (
        TOKEN_LIMIT,
        TOKEN_OFFSET,
        TOKEN_COLS,
    )
)

parse_range = make_parse_range(parse_simple_obj)


def parse_continue(_: TokenStream) -> Continue:
    """Parse the special `continue` value for the `offset` argument."""
    return CONTINUE


def parse_string_argument(stream: TokenStream) -> LoopArgument:
    """Try to parse a string literal as an integer."""
    try:
        return IntegerLiteral(to_int(stream.current[2]))
    except ValueError as err:
        raise LiquidSyntaxError(
            f"invalid integer argument {stream.current[2]!r}"
        ) from err


TOKEN_MAP: Dict[str, Callable[[TokenStream], LoopArgument]] = {
    TOKEN_IDENTIFIER: parse_identifier,
    TOKEN_INTEGER: parse_integer_literal,
    TOKEN_FLOAT: parse_float_literal,
    TOKEN_CONTINUE: parse_continue,
    TOKEN_STRING: parse_string_argument,
}


def parse_loop_argument(stream: TokenStream) -> LoopArgument:
    """Parse a object from the stream of tokens as a loop argument."""
    try:
        return TOKEN_MAP[stream.current[1]](stream)
    except KeyError as err:
        raise LiquidSyntaxError(f"unexpected {stream.current[2]!r}") from err


def parse_loop_arguments(stream: TokenStream) -> Tuple[Dict[str, LoopArgument], bool]:
    """Parse zero or more arguments from the stream of tokens until the end of the
    stream."""
    arguments: Dict[str, LoopArgument] = {}
    _reversed = False

    while True:
        _, typ, val = stream.current
        if typ == TOKEN_EOF:
            break

        if typ in LOOP_ARGS:
            next(stream)
            stream.expect(TOKEN_COLON)
            next(stream)
            arguments[val] = parse_loop_argument(stream)
            next(stream)
        elif typ == TOKEN_REVERSED:
            next(stream)
            _reversed = True
        elif typ == TOKEN_COMMA:
            next(stream)
        else:
            raise LiquidSyntaxError(f"unexpected {val!r}")

    return arguments, _reversed


def parse(expr: str, linenum: int = 1) -> LoopExpression:
    """Parse a loop expression string."""
    stream = TokenStream(tokenize(expr, linenum))
    stream.expect(TOKEN_IDENTIFIER)
    name = next(stream)[2]

    # Eat TOKEN_IN
    stream.expect(TOKEN_IN)
    next(stream)

    if stream.current[1] == TOKEN_IDENTIFIER:
        expression: LoopIterable = parse_identifier(stream)
        next(stream)
    elif stream.current[1] == TOKEN_LPAREN:
        expression = parse_range(stream)
        next(stream)
    else:
        raise LiquidSyntaxError("invalid loop expression", linenum=stream.current[0])

    args, reversed_ = parse_loop_arguments(stream)
    return LoopExpression(name=name, iterable=expression, reversed_=reversed_, **args)



"""Test cases for parsing loop-style expressions."""
import unittest

from typing import Any
from typing import Mapping
from typing import NamedTuple

from liquid import Context
from liquid import Environment
from liquid.expressions import parse_loop_expression


class Case(NamedTuple):
    """Table-driven test case helper."""

    description: str
    context: Mapping[str, Any]
    expression: str
    expect: Any


class EvalLoopExpressionTestCase(unittest.TestCase):
    """Test cases for evaluating `for` loop expressions."""

    def test_eval_loop_expression(self):
        """Test that we can evaluate loop expressions."""
        test_cases = [
            Case(
                description="simple range loop",
                context={},
                expression="i in (0..3)",
                expect=[0, 1, 2, 3],
            ),
            Case(
                description="reversed range",
                context={},
                expression="i in (0..3) reversed",
                expect=[3, 2, 1, 0],
            ),
            Case(
                description="loop over an object from context",
                context={"a": {"name": "foo", "title": "bar"}},
                expression="i in a",
                expect=[("name", "foo"), ("title", "bar")],
            ),
            Case(
                description="loop over nested and chained object from context",
                context={
                    "linklists": {"main": ["1", "2"]},
                    "section": {"settings": {"menu": "main"}},
                },
                expression="link in linklists[section.settings.menu]",
                expect=["1", "2"],
            ),
            Case(
                description=(
                    "loop over nested and chained object from context "
                    "with trailing identifier"
                ),
                context={
                    "linklists": {"main": {"links": ["1", "2"]}},
                    "section": {"settings": {"menu": "main"}},
                },
                expression="link in linklists[section.settings.menu].links",
                expect=["1", "2"],
            ),
            Case(
                description="chained identifier in range loop",
                context={
                    "foo": {"bar": 3},
                },
                expression="x in (1..foo.bar)",
                expect=[1, 2, 3],
            ),
        ]

        env = Environment()

        for case in test_cases:
            context = Context(env, case.context)
            with self.subTest(msg=case.description):
                expr = parse_loop_expression(case.expression)
                loop_iter, length = expr.evaluate(context)
                self.assertEqual(list(loop_iter), case.expect)
                self.assertEqual(length, len(case.expect))

    def test_eval_continue_loop_expression(self):
        """Test that we can evaluate loop expressions that use a continue offset."""
        env = Environment()
        context = Context(env, {"array": [1, 2, 3, 4, 5, 6]})

        # Mock a for loop with a limit
        context.stopindex("item-array", 3)
        expr = parse_loop_expression("item in array offset: continue")

        loop_iter, length = expr.evaluate(context)
        self.assertEqual(list(loop_iter), [4, 5, 6])
        self.assertEqual(length, 3)



"""Mock filter functions for displaying values as weights."""
# pylint: disable=missing-function-docstring
from liquid.filter import math_filter


@math_filter
def weight(grams):
    return f"{grams / 1000:.2f}"


@math_filter
def weight_with_unit(grams):
    return f"{grams / 1000:.2f} kg"



"""Golden tests cases for testing liquid's `liquid` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="newline terminated tags",
        template="\n".join(
            [
                r"{% liquid",
                r"if product.title",
                r"   echo product.title | upcase",
                r"else",
                r"   echo 'product-1' | upcase ",
                r"endif",
                r"",
                r"for i in (0..5)",
                r"   echo i",
                r"endfor %}",
            ]
        ),
        expect="FOO012345",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="carriage return and newline terminated tags",
        template="\r\n".join(
            [
                r"{% liquid",
                r"if product.title",
                r"   echo product.title | upcase",
                r"else",
                r"   echo 'product-1' | upcase ",
                r"endif",
                r"",
                r"for i in (0..5)",
                r"   echo i",
                r"endfor %}",
            ]
        ),
        expect="FOO012345",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="carriage return terminated tags",
        template="\r".join(
            [
                r"{% liquid",
                r"if product.title",
                r"   echo product.title | upcase",
                r"else",
                r"   echo 'product-1' | upcase ",
                r"endif",
                r"",
                r"for i in (0..5)",
                r"   echo i",
                r"endfor %}",
            ]
        ),
        expect="",
        globals={"product": {"title": "foo"}},
        error=True,
    ),
    Case(
        description="empty liquid tag",
        template=r"{% liquid %}",
        expect="",
    ),
    Case(
        description="only whitespace",
        template="{% liquid\n   \n\n   \t \n\t\n  %}",
        expect="",
    ),
    Case(
        description="single line comment tag",
        template="\n".join(
            [
                r"{% liquid",
                r"comment this is a comment",
                r"endcomment",
                r"%}",
            ]
        ),
        expect="",
    ),
    Case(
        description="multi-line comment tag",
        template="\n".join(
            [
                r"{% liquid",
                r"comment this is a comment",
                r"split over two lines",
                r"endcomment",
                r"%}",
            ]
        ),
        expect="",
    ),
    Case(
        description="whitespace control",
        template="\n".join(
            [
                r"Hello,     ",
                r"{%- liquid",
                r"  echo ' World! '",
                r"-%}",
                r"   Goodbye.",
            ]
        ),
        expect="Hello, World! Goodbye.",
    ),
    Case(
        description="reference test #2",
        template="\n".join(
            [
                r"{%- liquid",
                r"  for value in array",
                r"    echo value",
                r"    unless forloop.last",
                r"      echo '#'",
                r"    endunless",
                r"  endfor",
                r"-%}",
            ]
        ),
        expect="1#2#3",
        globals={"array": [1, 2, 3]},
    ),
    Case(
        description="reference test #3",
        template="\n".join(
            [
                r"{%- liquid",
                r"  for value in array",
                r"    assign double_value = value | times: 2",
                r"    echo double_value | times: 2",
                r"    unless forloop.last",
                r"      echo '#'",
                r"    endunless",
                r"  endfor",
                r"",
                r"  echo '#'",
                r"  echo double_value",
                r"-%}",
            ]
        ),
        expect="4#8#12#6",
        globals={"array": [1, 2, 3]},
    ),
    Case(
        description="reference test #4",
        template="\n".join(
            [
                r"{%- liquid echo 'a' -%}",
                r"b",
                r"{%- liquid echo 'c' -%}",
            ]
        ),
        expect="abc",
    ),
    Case(
        description="nested liquid",
        template="\n".join(
            [
                r"{%- if true %}",
                r"  {%- liquid",
                r'    echo "good"',
                r"  %}",
                r"{%- endif -%}",
            ]
        ),
        expect="good",
    ),
    Case(
        description="can't close nested blocks",
        template="\n".join(
            [
                r"{%- if true -%}",
                r"42",
                r"{%- liquid endif -%}",
            ]
        ),
        expect="",
        error=True,
    ),
]



"""Template output streams."""
from io import StringIO
from typing import Optional

from liquid.exceptions import OutputStreamLimitError


class LimitedStringIO(StringIO):
    """A StringIO subclass that limits the number of bytes that can be written."""

    def __init__(
        self,
        limit: int,
        initial_value: Optional[str] = None,
        newline: Optional[str] = None,
    ) -> None:
        super().__init__(initial_value, newline)
        self.limit = limit
        self.size = 0

    def write(self, __s: str) -> int:
        if __s:
            self.size += len(__s.encode("utf-8"))
            if self.size > self.limit:
                raise OutputStreamLimitError("output stream limit reached")
        return super().write(__s)



"""Golden tests cases for testing liquid's built-in `round` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="float round down",
        template=r"{{ 5.1 | round }}",
        expect="5",
    ),
    Case(
        description="float round up",
        template=r"{{ 5.6 | round }}",
        expect="6",
    ),
    Case(
        description="float as a string",
        template=r'{{ "5.6" | round }}',
        expect="6",
    ),
    Case(
        description="string argument",
        template=r'{{ 5.666 | round: "1" }}',
        expect="5.7",
    ),
    Case(
        description="decimal places",
        template=r'{{ "5.666666" | round: 2 }}',
        expect="5.67",
    ),
    Case(
        description="integer",
        template=r"{{ 5 | round }}",
        expect="5",
    ),
    Case(
        description="too many args",
        template=r"{{ 5 | round: 1, 2 }}",
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | round: 2 }}",
        expect="0",
    ),
    Case(
        description="undefined argument",
        template=r"{{ 5.666 | round: nosuchthing }}",
        expect="6.0",
    ),
]



"""Golden tests cases for testing liquid's `raw` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="literal",
        template=r"{% raw %}foo{% endraw %}",
        expect="foo",
    ),
    Case(
        description="statement",
        template=r"{% raw %}{{ foo }}{% endraw %}",
        expect=r"{{ foo }}",
    ),
    Case(
        description="tag",
        template=r"{% raw %}{% assign x = 1 %}{% endraw %}",
        expect=r"{% assign x = 1 %}",
    ),
    Case(
        description="partial tag",
        template=r"{% raw %} %} {% }} {{ {% endraw %}",
        expect=r" %} {% }} {{ ",
    ),
    Case(
        description="continue after raw",
        template=r"{% raw %} {% some raw content %} {% endraw %}a literal",
        expect=r" {% some raw content %} a literal",
    ),
]



"""Golden tests cases for testing liquid's `render` tag."""

from liquid.golden.case import Case
from liquid.golden.case import TEMPLATE_DROP_ATTRS

cases = [
    Case(
        description="string literal name",
        template=r"{% render 'product-hero', product: product %}",
        expect="foo\n- sports - garden ",
        globals={"product": {"title": "foo", "tags": ["sports", "garden"]}},
        partials={
            "product-hero": "\n".join(
                [
                    r"{{ product.title }}",
                    r"{% for tag in product.tags %}- {{ tag }} {% endfor %}",
                ]
            ),
        },
    ),
    Case(
        description="bound variable",
        template=r"{% render 'product-title' with collection.products[1] %}",
        expect="car",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={"product-title": r"{{ product-title.title }}"},
    ),
    Case(
        description="bound variable does not exist",
        template=r"{% render 'product-title' with no.such.thing %}",
        expect="",
        globals={},
        partials={"product-title": r"{{ product-title.title }}"},
    ),
    Case(
        description="bound array variable",
        template=r"{% render 'prod' for collection.products %}",
        expect="bikecar",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={"prod": r"{{ prod.title }}"},
    ),
    Case(
        description="bound variable with alias",
        template=r"{% render 'product-alias' with collection.products[1] as product %}",
        expect="car",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={
            "product-alias": r"{{ product.title }}",
        },
    ),
    Case(
        description="some keyword arguments",
        template=r"{% render 'product-args', foo: 'hello', bar: 'there' %}",
        expect="hello there",
        globals={},
        partials={
            "product-args": r"{{ foo }} {{ bar }}",
        },
    ),
    Case(
        description="some keyword arguments including a range literal",
        template=r"{% render 'product-args', foo: (1..3), bar: 'there' %}",
        expect="1#2#3 there",
        globals={},
        partials={
            "product-args": r"{{ foo | join: '#' }} {{ bar }}",
        },
    ),
    Case(
        description="some keyword arguments no leading coma",
        template=r"{% render 'product-args' foo: 'hello', bar: 'there' %}",
        expect="hello there",
        globals={},
        partials={"product-args": r"{{ foo }} {{ bar }}"},
    ),
    Case(
        description="template drop",
        template=r"{% render 'some/template-attrs.alt.txt' %}",
        expect="template-attrs.alt some template-attrs alt",
        globals={},
        partials={"some/template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no parent",
        template=r"{% render 'template-attrs.alt.txt' %}",
        expect="template-attrs.alt  template-attrs alt",
        globals={},
        partials={"template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no suffix",
        template=r"{% render 'some/template-attrs.txt' %}",
        expect="template-attrs some template-attrs ",
        globals={},
        partials={"some/template-attrs.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no suffix or extension",
        template=r"{% render 'some/template-attrs' %}",
        expect="template-attrs some template-attrs ",
        globals={},
        partials={"some/template-attrs": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="parent variables go out of scope",
        template=(
            r"{% assign greeting = 'good morning' %}"
            r"{{ greeting }} "
            r"{% render 'outer-scope' %}"
            r"{{ greeting }}"
        ),
        expect="good morning good morning",
        partials={"outer-scope": r"{{ greeting }}"},
    ),
    Case(
        description="for loop variables go out of scope",
        template=(
            r"{% for i in (1..3) %}"
            r"{{ i }}"
            r"{% render 'loop-scope' %}"
            r"{{ i }}"
            r"{% endfor %}"
            r"{{ i }}"
        ),
        expect="112233",
        partials={"loop-scope": r"{{ i }}"},
    ),
    Case(
        description="assigned variables do not leak into outer scope",
        template=(
            r"{% render 'assign-outer-scope', customer: customer %} {{ last_name }}"
        ),
        expect="Hello, Holly ",
        globals={"customer": {"first_name": "Holly"}},
        partials={
            "assign-outer-scope": (
                r"Hello, {{ customer.first_name }}{% assign last_name = 'Smith' %}"
            )
        },
    ),
    Case(
        description="increment is isolated between renders",
        template=r"{% increment foo %} {% render 'increment' %} {% increment foo %}",
        expect="0 0 1",
        partials={"increment": r"{% increment foo %}"},
    ),
    Case(
        description="decrement is isolated between renders",
        template=r"{% decrement foo %} {% render 'decrement' %} {% decrement foo %}",
        expect="-1 -1 -2",
        partials={"decrement": r"{% decrement foo %}"},
    ),
    Case(
        description="forloop helper",
        template=r"{% render 'product' for collection.products %}",
        expect="Product: bike first index:1 Product: car last index:2 ",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={
            "product": (
                r"Product: {{ product.title }} "
                r"{% if forloop.first %}first{% endif %}"
                r"{% if forloop.last %}last{% endif %}"
                r" index:{{ forloop.index }} "
            ),
        },
    ),
    Case(
        description="render loops don't add parentloop",
        template=r"{% render 'product' for collection.products %}",
        expect="bike-0 0 1 2 car-1 0 1 2 ",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={
            "product": (
                r"{{ product.title }}-{{ forloop.index0 }} "
                r"{% for x in (1..3) %}"
                r"{{ forloop.index0 }}{{ forloop.parentloop.index0 }} "
                r"{% endfor %}"
            ),
        },
    ),
    Case(
        description="render loops can't access parentloop",
        template=(
            r"{% for x in (1..3) %}"
            r"{% render 'product' for collection.products %}"
            r"{% endfor %}"
        ),
        expect="bike-0 car-1 bike-0 car-1 bike-0 car-1 ",
        globals={"collection": {"products": [{"title": "bike"}, {"title": "car"}]}},
        partials={
            "product": (
                r"{{ product.title }}-{{ forloop.index0 }} "
                r"{{ forloop.parentloop.index0 }}"
            ),
        },
    ),
    Case(
        description="assign to keyword argument",
        template=r"{% render 'product-args', foo: 'hello' %}{{ foo }}",
        expect="hello goodbye",
        globals={},
        partials={
            "product-args": r"{{ foo }}{% assign foo='goodbye' %} {{ foo }}",
        },
    ),
]



"""Golden tests cases for testing liquid whitespace control."""

from liquid.golden.case import Case

cases = [
    Case(
        description="white space control with newlines and spaces",
        template="".join(
            [
                "\n{% if customer -%}\n",
                "Welcome back,  {{ customer.first_name -}} !\n",
                " {%- endif -%}",
            ]
        ),
        expect="\nWelcome back,  Holly!",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description="white space control with carriage return and spaces",
        template="".join(
            [
                "\r{% if customer -%}\r",
                "Welcome back,  {{ customer.first_name -}} !\r",
                " {%- endif -%}",
            ]
        ),
        expect="\rWelcome back,  Holly!",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description=("white space control with  carriage return, newline and spaces"),
        template="".join(
            [
                "\r\n{% if customer -%}\r\n",
                "Welcome back,  {{ customer.first_name -}} !\r\n",
                " {%- endif -%}",
            ]
        ),
        expect="\r\nWelcome back,  Holly!",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description="white space control with newlines, tabs and spaces",
        template="".join(
            [
                "\n\t{% if customer -%}\t\n",
                "Welcome back,  {{ customer.first_name -}}\t !\r\n",
                " {%- endif -%}",
            ]
        ),
        expect="\n\tWelcome back,  Holly!",
        globals={"customer": {"first_name": "Holly"}},
    ),
    Case(
        description="suppress whitespace only if blocks",
        template="\n".join(
            [
                "!{% if true %}",
                "",
                "{% assign bar = 'foo' %}",
                "{% if true %}",
                "",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "{% endif %}",
                "",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!!",
        globals={},
    ),
    Case(
        description="suppress whitespace only unless blocks",
        template="\n".join(
            [
                "!{% unless false %}",
                "",
                "{% assign bar = 'foo' %}",
                "{% unless false %}",
                "",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "{% endunless %}",
                "",
                "",
                "{% endunless %}!",
            ]
        ),
        expect="!!",
        globals={},
    ),
    Case(
        description="suppress whitespace only case blocks",
        template="\n".join(
            [
                "!{% assign x = 1 %}{% case x %}",
                "",
                "  {% when 1 %}",
                "    {% assign foo = 'bar' %}",
                "",
                "",
                "{% endcase %}!",
            ]
        ),
        expect="!!",
        globals={},
    ),
    Case(
        description="don't suppress whitespace only blocks containing output",
        template="\n".join(
            [
                "!{% if true %}",
                "",
                "{% assign bar = 'foo' %}",
                "    {{ '' }}",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!\n\n\n    \n\n    \n\n\n\n!",
        globals={},
    ),
    Case(
        description="don't suppress whitespace only blocks containing echo",
        template="\n".join(
            [
                "!{% if true %}",
                "",
                "{% assign bar = 'foo' %}",
                "    {% echo '' %}",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!\n\n\n    \n\n    \n\n\n\n!",
        globals={},
    ),
    Case(
        description=(
            "don't suppress whitespace only blocks containing output in nested block"
        ),
        template="\n".join(
            [
                "!{% if 1 %}",
                "",
                "{% assign bar = 'foo' %}",
                "{% if 2 %}",
                "    {{ '' }}",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "{% endif %}",
                "",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!\n\n\n\n    \n\n    \n\n\n\n\n!",
        globals={},
    ),
    Case(
        description=(
            "don't suppress whitespace only unless blocks containing output in nested "
            "blocks"
        ),
        template="\n".join(
            [
                "!{% unless false %}",
                "",
                "{% assign bar = 'foo' %}",
                "{% unless false %}",
                "    {{ '' }}",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "{% endunless %}",
                "",
                "",
                "{% endunless %}!",
            ]
        ),
        expect="!\n\n\n\n    \n\n    \n\n\n\n\n!",
        globals={},
    ),
    Case(
        description=(
            "don't suppress whitespace only blocks containing output in "
            "unreachable blocks"
        ),
        template="\n".join(
            [
                "!{% if 1 %}",
                "",
                "{% assign bar = 'foo' %}",
                "{% if true %}",
                "",
                "    {% assign foo = 'bar' %}",
                "",
                "{% else %}",
                "    {{ '' }}",
                "{% endif %}",
                "",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!\n\n\n\n\n    \n\n\n\n\n!",
        globals={},
    ),
    Case(
        description="suppress whitespace surrounding an empty capture block",
        template="\n".join(
            [
                "!{% if true %}",
                "",
                "{% capture foo %}{% endcapture %}",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!!",
        globals={},
    ),
    Case(
        description="suppress whitespace surrounding a capture block",
        template="\n".join(
            [
                "!{% if true %}",
                "",
                "{% capture foo %}",
                "{{ '' }}",
                "{% endcapture %}",
                "",
                "{% endif %}!",
            ]
        ),
        expect="!!",
        globals={},
    ),
    Case(
        description="don't suppress whitespace only case blocks containing output",
        template="\n".join(
            [
                "!{% assign x = 1 %}{% case x %}",
                "",
                "  {% when 1 %}",
                "    {% assign foo = 'bar' %}",
                "",
                "  {% when 2 %}",
                "    {{ '' }}",
                "",
                "{% endcase %}!",
            ]
        ),
        expect="!\n    \n\n  !",
        globals={},
    ),
]



# pylint: disable=missing-class-docstring missing-module-docstring
import unittest

from dataclasses import dataclass
from typing import Union

from liquid.expression import Expression
from liquid.expression import IdentifierPathElement
from liquid.expression import Identifier
from liquid.expression import IntegerLiteral
from liquid.expression import RangeLiteral
from liquid.expression import LoopExpression

from liquid.expressions import parse_loop_expression


@dataclass
class Case:
    description: str
    expression: str
    expect: Union[Expression, str]


class ParseBooleanExpressionTestCase(unittest.TestCase):
    def test_parse_boolean_expression(self):
        """Test that we can parse liquid loop expressions."""
        test_cases = [
            Case(
                "item in array",
                "product in collection.products",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in array with integer literal limit",
                "product in collection.products limit:5",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                ),
            ),
            Case(
                "item in array with identifier limit",
                "product in collection.products limit:max",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=Identifier(path=[IdentifierPathElement("max")]),
                ),
            ),
            Case(
                "item in array with integer literal offset",
                "product in collection.products offset:2",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit",
                "product in collection.products limit:5 offset:2",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit out of order",
                "product in collection.products offset:2 limit:5",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array reversed",
                "product in collection.products reversed",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    reversed_=True,
                ),
            ),
            Case(
                "item in array with all options",
                "product in collection.products offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
            Case(
                "item in range with integer literal start and stop",
                "product in (1..10)",
                LoopExpression(
                    name="product",
                    iterable=RangeLiteral(
                        start=IntegerLiteral(1),
                        stop=IntegerLiteral(10),
                    ),
                ),
            ),
            Case(
                "item in range with integer literal start and identifier stop",
                "product in (1..collection.products.size)",
                LoopExpression(
                    name="product",
                    iterable=RangeLiteral(
                        start=IntegerLiteral(1),
                        stop=Identifier(
                            path=[
                                IdentifierPathElement("collection"),
                                IdentifierPathElement("products"),
                                IdentifierPathElement("size"),
                            ],
                        ),
                    ),
                ),
            ),
            Case(
                "item in range with integer identifier start and stop",
                "product in (num..collection.products.size)",
                LoopExpression(
                    name="product",
                    iterable=RangeLiteral(
                        start=Identifier(path=[IdentifierPathElement("num")]),
                        stop=Identifier(
                            path=[
                                IdentifierPathElement("collection"),
                                IdentifierPathElement("products"),
                                IdentifierPathElement("size"),
                            ],
                        ),
                    ),
                ),
            ),
            Case(
                "item in range with options",
                "product in (1..10) offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    iterable=RangeLiteral(
                        start=IntegerLiteral(1),
                        stop=IntegerLiteral(10),
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
            Case(
                "items in a nested array",
                "product in collections[0]['tags']",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collections"),
                            IdentifierPathElement(0),
                            IdentifierPathElement("tags"),
                        ],
                    ),
                ),
            ),
            Case(
                "split over multiple lines",
                "product\n in \ncollections[0]['tags']",
                LoopExpression(
                    name="product",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("collections"),
                            IdentifierPathElement(0),
                            IdentifierPathElement("tags"),
                        ],
                    ),
                ),
            ),
            Case(
                "comma separated arguments",
                "i in array, limit: 4, offset: 2",
                LoopExpression(
                    name="i",
                    iterable=Identifier(
                        path=[
                            IdentifierPathElement("array"),
                        ],
                    ),
                    limit=IntegerLiteral(4),
                    offset=IntegerLiteral(2),
                ),
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                expr = parse_loop_expression(case.expression)
                self.assertEqual(expr, case.expect)



"""Golden tests cases for testing liquid's built-in `append` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="concat",
        template=r'{{ "hello" | append: "there" }}',
        expect="hellothere",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | append: 'there' }}",
        expect="5there",
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello" | append: 5 }}',
        expect="hello5",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello" | append }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | append: "how", "are", "you" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | append: "hi" }}',
        expect="hi",
    ),
    Case(
        description="undefined argument",
        template=r'{{ "hi" | append: nosuchthing }}',
        expect="hi",
    ),
]



"""Test html filter functions."""
import unittest

from functools import partial
from inspect import isclass

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple

try:
    import markupsafe  # pylint: disable=unused-import # noqa: F401

    MARKUPSAFE_AVAILABLE = True
except ImportError:
    MARKUPSAFE_AVAILABLE = False

from liquid.environment import Environment
from liquid.exceptions import Error

from liquid.extra.filters.html import script_tag
from liquid.extra.filters.html import stylesheet_tag

from liquid import Markup


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class HTMLFilterTestCase(unittest.TestCase):
    """Test HTML filter functions."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases: Iterable[Case]):
        if getattr(func, "with_environment", False):
            func = partial(func, environment=self.env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(case.expect, Error):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_stylesheet_tag(self):
        """Test `stylesheet_tag` filter function."""
        test_cases = [
            Case(
                description="relative url",
                val="assets/style.css",
                args=[],
                kwargs={},
                expect=(
                    '<link href="assets/style.css" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
            Case(
                description="remote url",
                val="https://example.com/static/style.css",
                args=[],
                kwargs={},
                expect=(
                    '<link href="https://example.com/static/style.css" '
                    'rel="stylesheet" type="text/css" media="all" />'
                ),
            ),
            Case(
                description="html escape url",
                val="<b>assets/style.css</b>",
                args=[],
                kwargs={},
                expect=(
                    '<link href="&lt;b&gt;assets/style.css&lt;/b&gt;" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
            Case(
                description="not a string",
                val=42,
                args=[],
                kwargs={},
                expect=(
                    '<link href="42" rel="stylesheet" type="text/css" media="all" />'
                ),
            ),
        ]

        self._test(stylesheet_tag, test_cases)

    @unittest.skipIf(not MARKUPSAFE_AVAILABLE, "this test requires markupsafe")
    def test_stylesheet_tag_auto_escape(self):
        """Test `stylesheet_tag` filter function when autoescape is enabled."""
        test_cases = [
            Case(
                description="relative url",
                val="assets/style.css",
                args=[],
                kwargs={},
                expect=(
                    '<link href="assets/style.css" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
            Case(
                description="unsafe url from context",
                val="<b>assets/style.css</b>",
                args=[],
                kwargs={},
                expect=(
                    '<link href="&lt;b&gt;assets/style.css&lt;/b&gt;" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
            Case(
                description="safe url from context",
                val=Markup("<b>assets/style.css</b>"),
                args=[],
                kwargs={},
                expect=(
                    '<link href="&lt;b&gt;assets/style.css&lt;/b&gt;" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
        ]

        self.env.autoescape = True
        self._test(stylesheet_tag, test_cases)

    def test_script_tag(self):
        """Test `script_tag` filter function."""
        test_cases = [
            Case(
                description="relative url",
                val="assets/app.js",
                args=[],
                kwargs={},
                expect='<script src="assets/app.js" type="text/javascript"></script>',
            ),
            Case(
                description="remote url",
                val="https://example.com/static/assets/app.js",
                args=[],
                kwargs={},
                expect=(
                    "<script "
                    'src="https://example.com/static/assets/app.js" '
                    'type="text/javascript">'
                    "</script>"
                ),
            ),
            Case(
                description="not a string",
                val=42,
                args=[],
                kwargs={},
                expect='<script src="42" type="text/javascript"></script>',
            ),
        ]

        self._test(script_tag, test_cases)

    @unittest.skipIf(not MARKUPSAFE_AVAILABLE, "this test requires markupsafe")
    def test_script_tag_auto_escape(self):
        """Test `script_tag` filter function when autoescape is enabled."""
        test_cases = [
            Case(
                description="relative url",
                val="assets/assets/app.js",
                args=[],
                kwargs={},
                expect=(
                    '<script src="assets/assets/app.js" '
                    'type="text/javascript"></script>'
                ),
            ),
            Case(
                description="unsafe url from context",
                val="<b>assets/assets/app.js</b>",
                args=[],
                kwargs={},
                expect=(
                    '<script src="&lt;b&gt;assets/assets/app.js&lt;/b&gt;" '
                    'type="text/javascript"></script>'
                ),
            ),
            Case(
                description="safe url from context",
                val=Markup("<b>assets/assets/app.js</b>"),
                args=[],
                kwargs={},
                expect=(
                    '<script src="&lt;b&gt;assets/assets/app.js&lt;/b&gt;" '
                    'type="text/javascript"></script>'
                ),
            ),
        ]

        self.env.autoescape = True
        self._test(script_tag, test_cases)



"""Golden tests cases for testing liquid's built-in `where` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of hashes",
        template=(
            r"{% assign x = a | where: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,foo)(title,bar)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="array of hashes with equality test",
        template=(
            r"{% assign x = a | where: 'title', 'bar' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="array of hashes with a missing key",
        template=(
            r"{% assign x = a | where: 'title', 'bar' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)",
        globals={"a": [{"heading": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="left value is not an array",
        template=r"{{ a | where: 'title' }}",
        expect="",
        globals={"a": 123},
        error=True,
    ),
    Case(
        description="missing argument",
        template=r"{{ a | where }}",
        expect="",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r"{{ a | where: 'title', 'foo', 'bar' }}",
        expect="",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
        error=True,
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | where: 'title' }}",
        expect="",
    ),
    Case(
        description="first argument is undefined",
        template=r"{{ a | where: nosuchthing }}",
        expect="",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="second argument is undefined",
        template=(
            r"{% assign x = a | where: 'title', nosuchthing %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,foo)(title,bar)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="both arguments are undefined",
        template=r"{{ a | where: nosuchthing, nothing }}",
        expect="",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
    ),
    Case(
        description="value is false",
        template=(
            r"{% assign x =  a | where: 'b', false %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(b,false)",
        globals={"a": [{"b": False}, {"b": "bar"}, {"b": None}]},
    ),
    Case(
        description="value is explicit nil",
        template=(
            r"{% assign x =  a | where: 'b', nil %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(b,bar)",
        globals={"a": [{"b": False}, {"b": "bar"}, {"b": None}]},
    ),
]



"""System wide limits.

Ideally these limits would be configurable from a liquid.Environment, but our
template parser functions, filter decorators and expression objects don't have
access to an Environment. Changing that would require significant refactoring
resulting in too many breaking changes.
"""

import os
import sys

from typing import Any

from liquid.exceptions import LiquidValueError


_MIN_STR_INT = 640
_DEFAULT_MAX_STR_INT = 4300


def _init_int_max_str_digits() -> int:
    var = "LIQUIDINTMAXSTRDIGITS"
    env = os.environ.get(var)

    if env is None:
        if hasattr(sys, "get_int_max_str_digits"):
            return sys.get_int_max_str_digits()  # type: ignore

        var = "PYTHONINTMAXSTRDIGITS"
        env = os.environ.get(var)

    if env is None:
        return _DEFAULT_MAX_STR_INT

    if not env.isdigit():
        raise TypeError(
            f"{var}: invalid limit; must be >= " f"{_MIN_STR_INT} or 0 for unlimited"
        )

    max_digits = int(env)

    if max_digits == 0 or max_digits >= _MIN_STR_INT:
        return max_digits

    raise ValueError(
        f"{var}: invalid limit; must be >= {_MIN_STR_INT} or 0 for unlimited"
    )


MAX_STR_INT = _init_int_max_str_digits()


def to_int(val: Any) -> int:
    """Prevent DoS by very large str to int conversion.

    See https://github.com/python/cpython/issues/95778
    """
    if (
        isinstance(val, (str, bytes, bytearray))
        and MAX_STR_INT != 0
        and len(val) > MAX_STR_INT
    ):
        raise LiquidValueError(
            f"integer string conversion limit ({MAX_STR_INT}) reached: "
            f"value has {len(val)} digits"
        )
    return int(val)



"""Filters that don't exist in the reference implementation."""
from __future__ import annotations
from typing import TYPE_CHECKING

try:
    from markupsafe import Markup
except ImportError:
    from liquid.exceptions import Markup  # type: ignore

from liquid.filter import string_filter
from liquid.filter import with_environment

if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment


@with_environment
@string_filter
def safe(val: str, *, environment: Environment) -> str:
    """Stringify and mark as safe."""
    if environment.autoescape:
        return Markup(val)
    return val



"""Liquid tag base class."""
from __future__ import annotations

from abc import ABC, abstractmethod

from typing import TYPE_CHECKING

from liquid.ast import Node
from liquid.ast import IllegalNode
from liquid.exceptions import Error
from liquid.parse import eat_block
from liquid.stream import TokenStream

if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment


class Tag(ABC):
    """Base class for all built-in and custom template tags."""

    block = True
    name = ""
    end = ""

    def __init__(self, env: Environment):
        self.env = env

    def get_node(self, stream: TokenStream) -> Node:
        """Wraps `Tag.parse`, possibly returning an `IllegalNode`."""
        tok = stream.current
        try:
            return self.parse(stream)
        except Error as err:
            if not err.linenum:
                err.linenum = tok.linenum

            self.env.error(err)

            if self.block and hasattr(self, "end"):
                eat_block(stream, (self.end,))

            return IllegalNode(tok)

    @abstractmethod
    def parse(self, stream: TokenStream) -> Node:
        """Return a parse tree node by parsing tokens from the given stream."""
